from pandas import DataFrame

from arch.data.utility import load_file


def load() -> DataFrame:
    """
    Load the graduate school admissions data used in the examples

    Returns
    -------
    data : DataFrame
        Dataset containing GRE, GPA and class rank, and admission decision
    """
    return load_file(__file__, "binary.csv.gz")



from typing import Optional

import numpy as np
import pandas as pd
from scipy import stats
from statsmodels.iolib.summary import Summary
from statsmodels.iolib.table import SimpleTable
from statsmodels.regression.linear_model import RegressionResults

from arch.typing import ArrayLike1D, ArrayLike2D
from arch.unitroot._shared import (
    ResidualCointegrationTestResult,
    _check_cointegrating_regression,
    _cross_section,
)
from arch.unitroot.critical_values.engle_granger import (
    CV_PARAMETERS,
    LARGE_PARAMETERS,
    SMALL_PARAMETERS,
    TAU_MAX,
    TAU_MIN,
    TAU_STAR,
)
from arch.unitroot.unitroot import ADF, TREND_DESCRIPTION


def engle_granger(
    y: ArrayLike1D,
    x: ArrayLike2D,
    trend: str = "c",
    *,
    lags: Optional[int] = None,
    max_lags: Optional[int] = None,
    method: str = "bic",
) -> "EngleGrangerTestResults":
    r"""
    Test for cointegration within a set of time series.

    Parameters
    ----------
    y : array_like
        The left-hand-side variable in the cointegrating regression.
    x : array_like
        The right-hand-side variables in the cointegrating regression.
    trend : {"n","c","ct","ctt"}, default "c"
        Trend to include in the cointegrating regression. Trends are:

        * "n": No deterministic terms
        * "c": Constant
        * "ct": Constant and linear trend
        * "ctt": Constant, linear and quadratic trends
    lags : int, default None
        The number of lagged differences to include in the Augmented
        Dickey-Fuller test used on the residuals of the
    max_lags : int, default None
        The maximum number of lags to consider when using automatic
        lag-length in the Augmented Dickey-Fuller regression.
    method: {"aic", "bic", "tstat"}, default "bic"
        The method used to select the number of lags included in the
        Augmented Dickey-Fuller regression.

    Returns
    -------
    EngleGrangerTestResults
        Results of the Engle-Granger test.

    See Also
    --------
    arch.unitroot.ADF
        Augmented Dickey-Fuller testing.
    arch.unitroot.PhillipsPerron
        Phillips & Perron's unit root test.
    arch.unitroot.cointegration.phillips_ouliaris
        Phillips-Ouliaris tests of cointegration.

    Notes
    -----
    The model estimated is

    .. math::

       Y_t = X_t \beta + D_t \gamma + \epsilon_t

    where :math:`Z_t = [Y_t,X_t]` is being tested for cointegration.
    :math:`D_t` is a set of deterministic terms that may include a
    constant, a time trend or a quadratic time trend.

    The null hypothesis is that the series are not cointegrated.

    The test is implemented as an ADF of the estimated residuals from the
    cross-sectional regression using a set of critical values that is
    determined by the number of assumed stochastic trends when the null
    hypothesis is true.
    """
    setup = _check_cointegrating_regression(y, x, trend)
    xsection = _cross_section(setup.y, setup.x, setup.trend)
    resid = xsection.resid
    # Never pass in the trend here since only used in x-section
    adf = ADF(resid, lags, trend="n", max_lags=max_lags, method=method)
    stat = adf.stat
    nobs = resid.shape[0] - adf.lags - 1
    num_x = setup.x.shape[1]
    cv = engle_granger_cv(trend, num_x, nobs)
    pv = engle_granger_pval(stat, trend, num_x)
    return EngleGrangerTestResults(
        stat, pv, cv, order=num_x, adf=adf, xsection=xsection
    )


class EngleGrangerTestResults(ResidualCointegrationTestResult):
    """
    Results class for Engle-Granger cointegration tests.

    Parameters
    ----------
    stat : float
        The Engle-Granger test statistic.
    pvalue : float
        The pvalue of the Engle-Granger test statistic.
    crit_vals : Series
        The critical values of the Engle-Granger specific to the sample size
        and model dimension.
    null : str
        The null hypothesis.
    alternative : str
        The alternative hypothesis.
    trend : str
        The model's trend description.
    order : int
        The number of stochastic trends in the null distribution.
    adf : ADF
        The ADF instance used to perform the test and lag selection.
    xsection : RegressionResults
        The OLS results used in the cross-sectional regression.
    """

    def __init__(
        self,
        stat: float,
        pvalue: float,
        crit_vals: pd.Series,
        null: str = "No Cointegration",
        alternative: str = "Cointegration",
        trend: str = "c",
        order: int = 2,
        adf: Optional[ADF] = None,
        xsection: Optional[RegressionResults] = None,
    ) -> None:
        super().__init__(
            stat, pvalue, crit_vals, null, alternative, trend, order, xsection
        )
        self.name = "Engle-Granger Cointegration Test"
        assert adf is not None
        self._adf = adf
        self._additional_info.update(
            {
                "ADF Lag length": self.lags,
                "Trend": self.trend,
                "Estimated Root ρ (γ+1)": self.rho,
                "Distribution Order": self.distribution_order,
            }
        )

    @property
    def lags(self) -> int:
        """The number of lags used in the Augmented Dickey-Fuller regression."""
        return self._adf.lags

    @property
    def max_lags(self) -> Optional[int]:
        """The maximum number of lags used in the lag-length selection."""
        return self._adf.max_lags

    @property
    def rho(self) -> float:
        r"""
        The estimated coefficient in the Dickey-Fuller Test

        Returns
        -------
        float
            The coefficient.

        Notes
        -----
        The value returned is :math:`\hat{\rho}=\hat{\gamma}+1` from the ADF
        regression

        .. math::

            \Delta y_t = \gamma y_{t-1} + \sum_{i=1}^p \delta_i \Delta y_{t-i}
                         + \epsilon_t
        """

        return 1 + self._adf.regression.params[0]

    def summary(self) -> Summary:
        """Summary of test, containing statistic, p-value and critical values"""
        table_data = [
            ("Test Statistic", f"{self.stat:0.3f}"),
            ("P-value", f"{self.pvalue:0.3f}"),
            ("ADF Lag length", f"{self.lags:d}"),
            ("Estimated Root ρ (γ+1)", f"{self.rho:0.3f}"),
        ]
        title = self.name

        table = SimpleTable(
            table_data,
            stubs=None,
            title=title,
            colwidths=18,
            datatypes=[0, 1],
            data_aligns=("l", "r"),
        )

        smry = Summary()
        smry.tables.append(table)

        cv_string = "Critical Values: "
        for val in self.critical_values.keys():
            p = str(int(val)) + "%"
            cv_string += f"{self.critical_values[val]:0.2f}"
            cv_string += " (" + p + ")"
            cv_string += ", "
        # Remove trailing ,<space>
        cv_string = cv_string[:-2]

        extra_text = [
            "Trend: " + TREND_DESCRIPTION[self._trend],
            cv_string,
            "Null Hypothesis: " + self.null_hypothesis,
            "Alternative Hypothesis: " + self.alternative_hypothesis,
            "Distribution Order: " + str(self.distribution_order),
        ]

        smry.add_extra_txt(extra_text)
        return smry


def engle_granger_cv(trend: str, num_x: int, nobs: int) -> pd.Series:
    """
    Critical Values for Engle-Granger t-tests

    Parameters
    ----------
    trend : {"n", "c", "ct", "ctt"}
        The trend included in the model
    num_x : The number of cross-sectional regressors in the model.
        Must be between 1 and 12.
    nobs : int
        The number of observations in the time series.

    Returns
    -------
    Series
        The critical values for 1, 5 and 10%
    """
    trends = ("n", "c", "ct", "ctt")
    if trend not in trends:
        valid = ",".join(trends)
        raise ValueError(f"Trend must by one of: {valid}")
    if not 1 <= num_x <= 12:
        raise ValueError(
            "The number of cross-sectional variables must be between 1 and "
            "12 (inclusive)"
        )
    tbl = CV_PARAMETERS[trend]

    crit_vals = {}
    for size in (10, 5, 1):
        params = tbl[size][num_x]
        x = 1.0 / (nobs ** np.arange(4.0))
        crit_vals[size] = x @ params
    return pd.Series(crit_vals)


def engle_granger_pval(stat: float, trend: str, num_x: int) -> float:
    """
    Asymptotic P-values for Engle-Granger t-tests

    Parameters
    ----------
    stat : float
        The test statistic
    trend : {"n", "c", "ct", "ctt"}
        The trend included in the model
    num_x : The number of cross-sectional regressors in the model.
        Must be between 1 and 12.

    Returns
    -------
    Series
        The critical values for 1, 5 and 10%
    """
    trends = ("n", "c", "ct", "ctt")
    if trend not in trends:
        valid = ",".join(trends)
        raise ValueError(f"Trend must by one of: {valid}")
    if not 1 <= num_x <= 12:
        raise ValueError(
            "The number of cross-sectional variables must be between 1 and "
            "12 (inclusive)"
        )
    key = (trend, num_x)
    if stat > TAU_MAX[key]:
        return 1.0
    elif stat < TAU_MIN[key]:
        return 0.0
    if stat > TAU_STAR[key]:
        params = np.array(LARGE_PARAMETERS[key])
    else:
        params = np.array(SMALL_PARAMETERS[key])
    order = params.shape[0]
    x = stat ** np.arange(order)
    return stats.norm().cdf(params @ x)



# -*- coding: utf-8 -*-
"""
Copyright (c) 2018-2019, Martin Larralde
Copyright (c) 2015-2018, Daniel Greenfeld
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

* Redistributions of source code must retain the above copyright notice, this
list of conditions and the following disclaimer.

* Redistributions in binary form must reproduce the above copyright notice,
this list of conditions and the following disclaimer in the documentation
and/or other materials provided with the distribution.

* Neither the name of cached-property or property-cached nor the names of its
contributors may be used to endorse or promote products derived from this
software without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import asyncio
import functools
import threading
from time import time
from typing import Any, Callable, Mapping, Optional
import weakref

__author__ = "Martin Larralde"
__email__ = "martin.larralde@ens-paris-saclay.fr"
__license__ = "BSD"
__version__ = "1.6.3"


class cached_property(property):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Deleting the attribute resets the property.
    Source: https://github.com/bottlepy/bottle/commit/fa7733e075da0d790d809aa3d2f53071897e6f76
    """  # noqa

    _sentinel = object()

    _update_wrapper = functools.update_wrapper

    def __init__(self, func) -> None:
        self.cache: Mapping[str, Any] = weakref.WeakKeyDictionary()
        self.func: Callable[[], Any] = func
        self._update_wrapper(func)  # type: ignore

    def __get__(self, obj, cls):
        if obj is None:
            return self

        if asyncio and asyncio.iscoroutinefunction(self.func):
            return self._wrap_in_coroutine(obj)

        value = self.cache.get(obj, self._sentinel)
        if value is self._sentinel:
            value = self.cache[obj] = self.func(obj)

        return value

    def __set_name__(self, owner, name):
        self.__name__: str = name

    def __set__(self, obj, value):
        self.cache[obj] = value

    def __delete__(self, obj):
        del self.cache[obj]

    def _wrap_in_coroutine(self, obj):
        @functools.wraps(obj)
        @asyncio.coroutine
        def wrapper():
            value = self.cache.get(obj, self._sentinel)
            if value is self._sentinel:
                self.cache[obj] = value = asyncio.ensure_future(self.func(obj))
            return value

        return wrapper()


class threaded_cached_property(cached_property):
    """
    A cached_property version for use in environments where multiple threads
    might concurrently try to access the property.
    """

    def __init__(self, func) -> None:
        super(threaded_cached_property, self).__init__(func)
        self.lock: threading.RLock = threading.RLock()

    def __get__(self, obj, cls):
        if obj is None:
            return self
        with self.lock:
            return super(threaded_cached_property, self).__get__(obj, cls)

    def __set__(self, obj, value):
        with self.lock:
            super(threaded_cached_property, self).__set__(obj, value)

    def __delete__(self, obj):
        with self.lock:
            super(threaded_cached_property, self).__delete__(obj)


class cached_property_with_ttl(cached_property):
    """
    A property that is only computed once per instance and then replaces itself
    with an ordinary attribute. Setting the ttl to a number expresses how long
    the property will last before being timed out.
    """

    def __init__(self, ttl=None) -> None:
        if callable(ttl):
            func = ttl
            ttl = None
        else:
            func = None
        self.ttl: Optional[int] = ttl
        super(cached_property_with_ttl, self).__init__(func)

    def __call__(self, func):
        super(cached_property_with_ttl, self).__init__(func)
        return self

    def __get__(self, obj, cls):
        if obj is None:
            return self

        now = time()
        if obj in self.cache:
            value, last_updated = self.cache[obj]
            if not self.ttl or self.ttl > now - last_updated:
                return value

        value, _ = self.cache[obj] = (self.func(obj), now)
        return value

    def __set__(self, obj, value):
        super(cached_property_with_ttl, self).__set__(obj, (value, time()))


# Aliases to make cached_property_with_ttl easier to use
cached_property_ttl = cached_property_with_ttl
timed_cached_property = cached_property_with_ttl


class threaded_cached_property_with_ttl(
    cached_property_with_ttl, threaded_cached_property
):
    """
    A cached_property version for use in environments where multiple threads
    might concurrently try to access the property.
    """

    def __init__(self, ttl=None) -> None:
        super(threaded_cached_property_with_ttl, self).__init__(ttl)
        self.lock: threading.RLock = threading.RLock()

    def __get__(self, obj, cls):
        with self.lock:
            return super(threaded_cached_property_with_ttl, self).__get__(obj, cls)

    def __set__(self, obj, value):
        with self.lock:
            return super(threaded_cached_property_with_ttl, self).__set__(obj, value)

    def __delete__(self, obj):
        with self.lock:
            return super(threaded_cached_property_with_ttl, self).__delete__(obj)


# Alias to make threaded_cached_property_with_ttl easier to use
threaded_cached_property_ttl = threaded_cached_property_with_ttl
timed_threaded_cached_property = threaded_cached_property_with_ttl



from distutils.version import LooseVersion
from io import StringIO
import sys
import warnings

import numpy as np
from numpy.random import RandomState
from numpy.testing import assert_almost_equal, assert_array_almost_equal, assert_equal
import pandas as pd
from pandas.testing import assert_frame_equal, assert_series_equal
import pytest
import scipy
from scipy import stats
from scipy.optimize import OptimizeResult
import statsmodels.regression.linear_model as smlm
import statsmodels.tools as smtools

from arch.univariate.base import ARCHModelForecast, ARCHModelResult, _align_forecast
from arch.univariate.distribution import (
    GeneralizedError,
    Normal,
    SkewStudent,
    StudentsT,
)
from arch.univariate.mean import ARX, HARX, LS, ConstantMean, ZeroMean, arch_model
from arch.univariate.volatility import (
    ARCH,
    EGARCH,
    FIGARCH,
    GARCH,
    HARCH,
    ConstantVariance,
    EWMAVariance,
    FixedVariance,
    MIDASHyperbolic,
    RiskMetrics2006,
)
from arch.utility.exceptions import ConvergenceWarning, DataScaleWarning

USE_CYTHON = False
try:
    import arch.univariate.recursions

    USE_CYTHON = True
except ImportError:
    import arch.univariate.recursions_python  # noqa

if USE_CYTHON:
    rec = arch.univariate.recursions
else:
    rec = arch.univariate.recursions_python

try:
    import matplotlib.pyplot  # noqa

    HAS_MATPLOTLIB = True
except ImportError:
    HAS_MATPLOTLIB = False

DISPLAY = "off"
SP_LT_14 = LooseVersion(scipy.__version__) < LooseVersion("1.4")


@pytest.fixture(scope="module", params=[True, False])
def simulated_data(request):
    rs = np.random.RandomState(1)
    zm = ZeroMean(volatility=GARCH(), distribution=Normal(rs))
    sim_data = zm.simulate(np.array([0.1, 0.1, 0.88]), 1000)
    return np.asarray(sim_data.data) if request.param else sim_data.data


class TestMeanModel(object):
    @classmethod
    def setup_class(cls):
        cls.rng = RandomState(1234)
        cls.T = 1000
        cls.resids = cls.rng.standard_normal(cls.T)
        zm = ZeroMean()
        zm.volatility = GARCH()
        seed = 12345
        random_state = np.random.RandomState(seed)
        zm.distribution = Normal(random_state=random_state)
        sim_data = zm.simulate(np.array([0.1, 0.1, 0.8]), 1000)
        with pytest.raises(ValueError):
            zm.simulate(np.array([0.1, 0.1, 0.8]), 1000, initial_value=3.0)
        date_index = pd.date_range("2000-12-31", periods=1000, freq="W")
        cls.y = sim_data.data.values
        cls.y_df = pd.DataFrame(
            cls.y[:, None], columns=["LongVariableName"], index=date_index
        )

        cls.y_series = pd.Series(
            cls.y, name="VeryVeryLongLongVariableName", index=date_index
        )
        x = cls.resids + cls.rng.standard_normal(cls.T)
        cls.x = x[:, None]
        cls.x_df = pd.DataFrame(cls.x, columns=["LongExogenousName"])
        cls.resid_var = np.var(cls.resids)
        cls.sigma2 = np.zeros_like(cls.resids)
        cls.backcast = 1.0

    def test_constant_mean(self):
        cm = ConstantMean(self.y)
        parameters = np.array([5.0, 1.0])
        cm.simulate(parameters, self.T)
        assert_equal(cm.num_params, 1)
        with pytest.raises(ValueError):
            cm.simulate(parameters, self.T, x=np.array(10))
        bounds = cm.bounds()
        assert_equal(bounds, [(-np.inf, np.inf)])
        assert_equal(cm.constant, True)
        a, b = cm.constraints()
        assert_equal(a, np.empty((0, 1)))
        assert_equal(b, np.empty((0,)))
        assert isinstance(cm.volatility, ConstantVariance)
        assert isinstance(cm.distribution, Normal)
        assert cm.lags is None
        res = cm.fit(disp=DISPLAY)
        expected = np.array([self.y.mean(), self.y.var()])
        assert_almost_equal(res.params, expected)

        forecasts = res.forecast(horizon=20, start=20)
        direct = pd.DataFrame(
            index=np.arange(self.y.shape[0]),
            columns=["h.{0:>02d}".format(i + 1) for i in range(20)],
            dtype="double",
        )
        direct.iloc[20:, :] = res.params.iloc[0]
        # TODO
        # assert_frame_equal(direct, forecasts)
        assert isinstance(forecasts, ARCHModelForecast)
        assert isinstance(cm.__repr__(), str)
        assert isinstance(cm.__str__(), str)
        assert "<strong>" in cm._repr_html_()
        with pytest.raises(ValueError, match="horizon must be an integer >= 1"):
            res.forecast(horizon=0, start=20)

    def test_zero_mean(self):
        zm = ZeroMean(self.y)
        parameters = np.array([1.0])
        data = zm.simulate(parameters, self.T)
        assert_equal(data.shape, (self.T, 3))
        assert_equal(data["data"].shape[0], self.T)
        assert_equal(zm.num_params, 0)
        bounds = zm.bounds()
        assert_equal(bounds, [])
        assert_equal(zm.constant, False)
        a, b = zm.constraints()
        assert_equal(a, np.empty((0, 0)))
        assert_equal(b, np.empty((0,)))
        assert isinstance(zm.volatility, ConstantVariance)
        assert isinstance(zm.distribution, Normal)
        assert zm.lags is None
        res = zm.fit(disp=DISPLAY)
        assert_almost_equal(res.params, np.array([np.mean(self.y ** 2)]))

        forecasts = res.forecast(horizon=99)
        direct = pd.DataFrame(
            index=np.arange(self.y.shape[0]),
            columns=["h.{0:>02d}".format(i + 1) for i in range(99)],
            dtype="double",
        )
        direct.iloc[:, :] = 0.0
        assert isinstance(forecasts, ARCHModelForecast)
        # TODO
        # assert_frame_equal(direct, forecasts)
        garch = GARCH()
        zm.volatility = garch
        zm.fit(update_freq=0, disp=DISPLAY)
        assert isinstance(zm.__repr__(), str)
        assert isinstance(zm.__str__(), str)
        assert "<strong>" in zm._repr_html_()

    def test_harx(self):
        harx = HARX(self.y, self.x, lags=[1, 5, 22])
        assert harx.x is self.x
        params = np.array([1.0, 0.4, 0.3, 0.2, 1.0, 1.0])
        harx.simulate(params, self.T, x=self.rng.randn(self.T + 500, 1))
        iv = self.rng.randn(22, 1)
        data = harx.simulate(
            params, self.T, x=self.rng.randn(self.T + 500, 1), initial_value=iv
        )
        assert_equal(data.shape, (self.T, 3))
        cols = ["data", "volatility", "errors"]
        for c in cols:
            assert c in data

        bounds = harx.bounds()
        for b in bounds:
            assert_equal(b[0], -np.inf)
            assert_equal(b[1], np.inf)
        assert_equal(len(bounds), 5)

        assert_equal(harx.num_params, 1 + 3 + self.x.shape[1])
        assert_equal(harx.constant, True)
        a, b = harx.constraints()
        assert_equal(a, np.empty((0, 5)))
        assert_equal(b, np.empty(0))
        res = harx.fit(disp=DISPLAY)
        with pytest.raises(ValueError):
            res.forecast(params=np.array([1.0, 1.0]))
        nobs = self.T - 22
        rhs = np.ones((nobs, 5))
        y = self.y
        lhs = y[22:]
        for i in range(self.T - 22):
            rhs[i, 1] = y[i + 21]
            rhs[i, 2] = np.mean(y[i + 17 : i + 22])
            rhs[i, 3] = np.mean(y[i : i + 22])
        rhs[:, 4] = self.x[22:, 0]
        params = np.linalg.pinv(rhs).dot(lhs)
        assert_almost_equal(params, res.params[:-1])

        assert harx.hold_back is None
        assert_equal(harx.lags, [1, 5, 22])
        assert_equal(harx.name, "HAR-X")
        assert_equal(harx.use_rotated, False)
        assert isinstance(harx.__repr__(), str)
        harx._repr_html_()
        res = harx.fit(cov_type="classic", disp=DISPLAY)
        assert isinstance(res.__repr__(), str)

    def test_harx_error(self):
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[1, -5, 22])
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[0, 1, 5, 22])
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[[-1], [3]])
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[[0], [0]])
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[[1, 1, 3], [2, 3, 3]])
        with pytest.raises(ValueError):
            HARX(self.y, self.x, lags=[[[1], [3]]])

    def test_har(self):
        har = HARX(self.y, lags=[1, 5, 22])
        params = np.array([1.0, 0.4, 0.3, 0.2, 1.0])
        data = har.simulate(params, self.T)
        assert_equal(data.shape, (self.T, 3))
        cols = ["data", "volatility", "errors"]
        for c in cols:
            assert c in data

        bounds = har.bounds()
        for b in bounds:
            assert_equal(b[0], -np.inf)
            assert_equal(b[1], np.inf)
        assert_equal(len(bounds), 4)

        assert_equal(har.num_params, 4)
        assert_equal(har.constant, True)
        a, b = har.constraints()
        assert_equal(a, np.empty((0, 4)))
        assert_equal(b, np.empty(0))
        res = har.fit(disp=DISPLAY)
        nobs = self.T - 22
        rhs = np.ones((nobs, 4))
        y = self.y
        lhs = y[22:]
        for i in range(self.T - 22):
            rhs[i, 1] = y[i + 21]
            rhs[i, 2] = np.mean(y[i + 17 : i + 22])
            rhs[i, 3] = np.mean(y[i : i + 22])
        params = np.linalg.pinv(rhs).dot(lhs)
        assert_almost_equal(params, res.params[:-1])

        with pytest.raises(ValueError):
            res.forecast(horizon=6, start=0)
        forecasts = res.forecast(horizon=6)
        t = self.y.shape[0]
        direct = pd.DataFrame(
            index=np.arange(t),
            columns=["h." + str(i + 1) for i in range(6)],
            dtype="float64",
        )

        params = np.asarray(res.params)
        fcast = np.zeros(t + 6)
        for i in range(21, t):
            fcast[: i + 1] = self.y[: i + 1]
            fcast[i + 1 :] = 0.0
            for h in range(6):
                fcast[i + h + 1] = params[0]
                fcast[i + h + 1] += params[1] * fcast[i + h : i + h + 1]
                fcast[i + h + 1] += params[2] * fcast[i + h - 4 : i + h + 1].mean()
                fcast[i + h + 1] += params[3] * fcast[i + h - 21 : i + h + 1].mean()
            direct.iloc[i, :] = fcast[i + 1 : i + 7]
        assert isinstance(forecasts, ARCHModelForecast)
        # TODO
        # assert_frame_equal(direct, forecasts)
        forecasts = res.forecast(res.params, horizon=6)
        assert isinstance(forecasts, ARCHModelForecast)
        # TODO
        # assert_frame_equal(direct, forecasts)

        assert har.hold_back is None
        assert_equal(har.lags, [1, 5, 22])
        assert_equal(har.name, "HAR")
        assert_equal(har.use_rotated, False)

        har = HARX(self.y_series, lags=[1, 5, 22])
        res = har.fit(disp=DISPLAY)
        direct = pd.DataFrame(
            index=self.y_series.index,
            columns=["h." + str(i + 1) for i in range(6)],
            dtype="float64",
        )
        forecasts = res.forecast(horizon=6)
        params = np.asarray(res.params)
        fcast = np.zeros(t + 6)
        for i in range(21, t):
            fcast[: i + 1] = self.y[: i + 1]
            fcast[i + 1 :] = 0.0
            for h in range(6):
                fcast[i + h + 1] = params[0]
                fcast[i + h + 1] += params[1] * fcast[i + h : i + h + 1]
                fcast[i + h + 1] += params[2] * fcast[i + h - 4 : i + h + 1].mean()
                fcast[i + h + 1] += params[3] * fcast[i + h - 21 : i + h + 1].mean()
            direct.iloc[i, :] = fcast[i + 1 : i + 7]
        assert isinstance(forecasts, ARCHModelForecast)
        # TODO
        # assert_frame_equal(direct, forecasts)

    def test_arx(self):
        arx = ARX(self.y, self.x, lags=3, hold_back=10, constant=False)
        params = np.array([0.4, 0.3, 0.2, 1.0, 1.0])
        data = arx.simulate(params, self.T, x=self.rng.randn(self.T + 500, 1))
        assert isinstance(data, pd.DataFrame)
        bounds = arx.bounds()
        for b in bounds:
            assert_equal(b[0], -np.inf)
            assert_equal(b[1], np.inf)
        assert_equal(len(bounds), 4)

        assert_equal(arx.num_params, 4)
        assert not arx.constant
        a, b = arx.constraints()
        assert_equal(a, np.empty((0, 4)))
        assert_equal(b, np.empty(0))
        res = arx.fit(last_obs=900, disp=DISPLAY)
        assert res.fit_stop == 900

        nobs = 900 - 10
        rhs = np.zeros((nobs, 4))
        y = self.y
        lhs = y[10:900]
        for i in range(10, 900):
            rhs[i - 10, 0] = y[i - 1]
            rhs[i - 10, 1] = y[i - 2]
            rhs[i - 10, 2] = y[i - 3]
        rhs[:, 3] = self.x[10:900, 0]
        params = np.linalg.pinv(rhs).dot(lhs)
        assert_almost_equal(params, res.params[:-1])
        assert_equal(arx.hold_back, 10)
        assert_equal(arx.lags, np.array([[1, 2, 3], [1, 2, 3]]))
        assert_equal(arx.name, "AR-X")
        assert_equal(arx.use_rotated, False)
        assert isinstance(arx.__repr__(), str)
        arx._repr_html_()

    def test_ar(self):
        ar = ARX(self.y, lags=3)
        params = np.array([1.0, 0.4, 0.3, 0.2, 1.0])
        data = ar.simulate(params, self.T)
        assert len(data) == self.T
        assert_equal(self.y, ar.y)

        bounds = ar.bounds()
        for b in bounds:
            assert_equal(b[0], -np.inf)
            assert_equal(b[1], np.inf)
        assert_equal(len(bounds), 4)

        assert_equal(ar.num_params, 4)
        assert ar.constant
        a, b = ar.constraints()
        assert_equal(a, np.empty((0, 4)))
        assert_equal(b, np.empty(0))
        res = ar.fit(disp=DISPLAY)

        nobs = 1000 - 3
        rhs = np.ones((nobs, 4))
        y = self.y
        lhs = y[3:1000]
        for i in range(3, 1000):
            rhs[i - 3, 1] = y[i - 1]
            rhs[i - 3, 2] = y[i - 2]
            rhs[i - 3, 3] = y[i - 3]
        params = np.linalg.pinv(rhs).dot(lhs)
        assert_almost_equal(params, res.params[:-1])

        forecasts = res.forecast(horizon=5)
        direct = pd.DataFrame(
            index=np.arange(y.shape[0]),
            columns=["h." + str(i + 1) for i in range(5)],
            dtype="float64",
        )
        params = res.params.iloc[:-1]
        for i in range(2, y.shape[0]):
            fcast = np.zeros(y.shape[0] + 5)
            fcast[: y.shape[0]] = y.copy()
            for h in range(1, 6):
                reg = np.array(
                    [1.0, fcast[i + h - 1], fcast[i + h - 2], fcast[i + h - 3]]
                )
                fcast[i + h] = reg.dot(params)
            direct.iloc[i, :] = fcast[i + 1 : i + 6]
        assert isinstance(forecasts, ARCHModelForecast)
        # TODO
        # assert_frame_equal(direct, forecasts)

        assert ar.hold_back is None
        assert_equal(ar.lags, np.array([[1, 2, 3], [1, 2, 3]]))
        assert_equal(ar.name, "AR")
        assert_equal(ar.use_rotated, False)
        ar.__repr__()
        ar._repr_html_()

        ar = ARX(self.y_df, lags=5)
        ar.__repr__()
        ar = ARX(self.y_series, lags=5)
        ar.__repr__()
        res = ar.fit(disp=DISPLAY)
        assert isinstance(res.resid, pd.Series)
        assert isinstance(res.conditional_volatility, pd.Series)
        std_resid = res.resid / res.conditional_volatility
        std_resid.name = "std_resid"
        assert_series_equal(res.std_resid, std_resid)
        # Smoke bootstrap
        summ = ar.fit(disp=DISPLAY).summary()
        assert "Df Model:                            6" in str(summ)
        assert "Constant Variance" in str(summ)
        ar = ARX(self.y, lags=1, volatility=GARCH(), distribution=StudentsT())
        res = ar.fit(disp=DISPLAY, update_freq=5, cov_type="classic")
        assert isinstance(res.param_cov, pd.DataFrame)
        sims = res.forecast(horizon=5, method="simulation")
        assert isinstance(sims.simulations.residual_variances, np.ndarray)
        assert isinstance(sims.simulations.residuals, np.ndarray)
        assert isinstance(sims.simulations.values, np.ndarray)
        assert isinstance(sims.simulations.variances, np.ndarray)

    def test_ar_no_lags(self):
        ar = ARX(self.y, lags=0)
        assert ar.lags is None
        res = ar.fit(disp=DISPLAY)
        assert_almost_equal(res.params[0], self.y.mean())
        assert "lags: none" in ar.__str__()

    @pytest.mark.skipif(not HAS_MATPLOTLIB, reason="matplotlib not installed")
    def test_ar_plot(self):
        ar = ARX(self.y, lags=1, volatility=GARCH(), distribution=StudentsT())
        res = ar.fit(disp=DISPLAY, update_freq=5, cov_type="mle")
        res.plot()
        res.plot(annualize="D")
        res.plot(annualize="W")
        res.plot(annualize="M")
        with pytest.raises(ValueError):
            res.plot(annualize="unknown")

        import matplotlib.pyplot as plt

        plt.close("all")

        res.plot(scale=360)
        res.hedgehog_plot(start=500)
        res.hedgehog_plot(start=500, plot_type="mean")
        res.hedgehog_plot(plot_type="volatility")
        res.hedgehog_plot(start=500, method="simulation", simulations=100)
        res.hedgehog_plot(plot_type="volatility", method="bootstrap")

        plt.close("all")

    def test_arch_arx(self):
        self.rng.seed(12345)
        x = self.rng.randn(500, 3)
        y = x.sum(1) + 3 * self.rng.standard_normal(500)

        am = ARX(y=y, x=x)
        res = am.fit(disp=DISPLAY)
        res.summary()
        assert isinstance(res.optimization_result, OptimizeResult)
        am.volatility = ARCH(p=2)
        results = am.fit(update_freq=0, disp=DISPLAY)
        assert isinstance(results.pvalues, pd.Series)
        assert_equal(
            list(results.pvalues.index),
            ["Const", "x0", "x1", "x2", "omega", "alpha[1]", "alpha[2]"],
        )

        am = ARX(y=y, lags=2, x=x)
        res = am.fit(disp=DISPLAY)
        summ = res.summary().as_text()
        res_repr = res.__repr__()
        assert str(hex(id(res))) in res_repr
        assert summ[:10] == res_repr[:10]

        am.volatility = ARCH(p=2)
        results = am.fit(update_freq=0, disp=DISPLAY)
        assert isinstance(results.pvalues, pd.Series)
        assert_equal(
            list(results.pvalues.index),
            [
                "Const",
                "y[1]",
                "y[2]",
                "x0",
                "x1",
                "x2",
                "omega",
                "alpha[1]",
                "alpha[2]",
            ],
        )

        x = pd.DataFrame(x, columns=["x0", "x1", "x2"])
        y = pd.Series(y, name="y")
        am = ARX(y=y, x=x)
        am.fit(disp=DISPLAY).summary()
        am.volatility = ARCH(p=2)
        results = am.fit(update_freq=0, disp=DISPLAY)
        assert isinstance(results.pvalues, pd.Series)
        assert_equal(
            list(results.pvalues.index),
            ["Const", "x0", "x1", "x2", "omega", "alpha[1]", "alpha[2]"],
        )

    def test_arch_model(self):
        am = arch_model(self.y)
        assert isinstance(am, ConstantMean)
        assert isinstance(am.volatility, GARCH)
        assert isinstance(am.distribution, Normal)

        am = arch_model(self.y, mean="harx", lags=[1, 5, 22])
        assert isinstance(am, HARX)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, mean="har", lags=[1, 5, 22])
        assert isinstance(am, HARX)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, self.x, mean="ls")
        assert isinstance(am, LS)
        assert isinstance(am.volatility, GARCH)
        am.__repr__()

        am = arch_model(self.y, mean="arx", lags=[1, 5, 22])
        assert isinstance(am, ARX)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, mean="ar", lags=[1, 5, 22])
        assert isinstance(am, ARX)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, mean="ar", lags=None)
        assert isinstance(am, ARX)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, mean="zero")
        assert isinstance(am, ZeroMean)
        assert isinstance(am.volatility, GARCH)

        am = arch_model(self.y, vol="Harch")
        assert isinstance(am, ConstantMean)
        assert isinstance(am.volatility, HARCH)

        am = arch_model(self.y, vol="Constant")
        assert isinstance(am, ConstantMean)
        assert isinstance(am.volatility, ConstantVariance)

        am = arch_model(self.y, vol="arch")
        assert isinstance(am.volatility, ARCH)

        am = arch_model(self.y, vol="egarch")
        assert isinstance(am.volatility, EGARCH)

        am = arch_model(self.y, vol="figarch")
        assert isinstance(am.volatility, FIGARCH)

        with pytest.raises(ValueError):
            arch_model(self.y, mean="unknown")
        with pytest.raises(ValueError):
            arch_model(self.y, vol="unknown")
        with pytest.raises(ValueError):
            arch_model(self.y, dist="unknown")

        am.fit(disp=DISPLAY)

    def test_pandas(self):
        am = arch_model(self.y_df, self.x_df, mean="ls")
        assert isinstance(am, LS)

    def test_summary(self):
        am = arch_model(self.y, mean="ar", lags=[1, 3, 5])
        res = am.fit(update_freq=0, disp=DISPLAY)
        res.summary()

        am = arch_model(self.y, mean="ar", lags=[1, 3, 5], dist="studentst")
        assert isinstance(am.distribution, StudentsT)
        res = am.fit(update_freq=0, disp=DISPLAY)
        res.summary()

        am = arch_model(self.y, mean="ar", lags=[1, 3, 5], dist="ged")
        assert isinstance(am.distribution, GeneralizedError)
        res = am.fit(update_freq=0, disp=DISPLAY)
        res.summary()

        am = arch_model(self.y, mean="ar", lags=[1, 3, 5], dist="skewt")
        res = am.fit(update_freq=0, disp=DISPLAY)
        assert isinstance(am.distribution, SkewStudent)
        res.summary()

    def test_errors(self):
        with pytest.raises(ValueError):
            ARX(self.y, lags=np.array([[1, 2], [3, 4]]))
        x = self.rng.randn(self.y.shape[0] + 1, 1)
        with pytest.raises(ValueError):
            ARX(self.y, x=x)
        with pytest.raises(ValueError):
            HARX(self.y, lags=np.eye(3))
        with pytest.raises(ValueError):
            ARX(self.y, lags=-1)
        with pytest.raises(ValueError):
            ARX(self.y, x=self.rng.randn(1, 1), lags=-1)

        ar = ARX(self.y, lags=1)
        with pytest.raises(ValueError):
            d = Normal()
            ar.volatility = d

        with pytest.raises(ValueError):
            v = GARCH()
            ar.distribution = v
        x = self.rng.randn(1000, 1)
        with pytest.raises(ValueError):
            ar.simulate(np.ones(5), 100, x=x)
        with pytest.raises(ValueError):
            ar.simulate(np.ones(5), 100)
        with pytest.raises(ValueError):
            ar.simulate(np.ones(3), 100, initial_value=self.rng.standard_normal(10))

        with pytest.raises(ValueError):
            ar.volatility = ConstantVariance()
            ar.fit(cov_type="unknown")

    def test_warnings(self):
        with warnings.catch_warnings(record=True) as w:
            ARX(self.y, lags=[1, 2, 3, 12], hold_back=5)
            assert_equal(len(w), 1)

        with warnings.catch_warnings(record=True) as w:
            HARX(self.y, lags=[[1, 1, 1], [2, 5, 22]], use_rotated=True)
            assert_equal(len(w), 1)

    def test_har_lag_specifications(self):
        """ Test equivalence of alternative lag specifications"""
        har = HARX(self.y, lags=[1, 2, 3])
        har_r = HARX(self.y, lags=[1, 2, 3], use_rotated=True)
        har_r_v2 = HARX(self.y, lags=3, use_rotated=True)
        ar = ARX(self.y, lags=[1, 2, 3])
        ar_v2 = ARX(self.y, lags=3)

        res_har = har.fit(disp=DISPLAY)
        res_har_r = har_r.fit(disp=DISPLAY)
        res_har_r_v2 = har_r_v2.fit(disp=DISPLAY)
        res_ar = ar.fit(disp=DISPLAY)
        res_ar_v2 = ar_v2.fit(disp=DISPLAY)
        assert_almost_equal(res_har.rsquared, res_har_r.rsquared)
        assert_almost_equal(res_har_r_v2.rsquared, res_har_r.rsquared)
        assert_almost_equal(np.asarray(res_ar.params), np.asarray(res_ar_v2.params))
        assert_almost_equal(np.asarray(res_ar.params), np.asarray(res_har_r_v2.params))
        assert_almost_equal(
            np.asarray(res_ar.param_cov), np.asarray(res_har_r_v2.param_cov)
        )
        assert_almost_equal(
            res_ar.conditional_volatility, res_har_r_v2.conditional_volatility
        )
        assert_almost_equal(res_ar.resid, res_har_r_v2.resid)

    def test_starting_values(self):
        am = arch_model(self.y, mean="ar", lags=[1, 3, 5])
        res = am.fit(cov_type="classic", update_freq=0, disp=DISPLAY)
        res2 = am.fit(starting_values=res.params, update_freq=0, disp=DISPLAY)
        assert isinstance(res, ARCHModelResult)
        assert isinstance(res2, ARCHModelResult)
        assert len(res.params) == 7
        assert len(res2.params) == 7

        am = arch_model(self.y, mean="zero")
        sv = np.array([1.0, 0.3, 0.8])
        with warnings.catch_warnings(record=True) as w:
            am.fit(starting_values=sv, update_freq=0, disp=DISPLAY)
            assert_equal(len(w), 1)

    def test_no_param_volatility(self):
        cm = ConstantMean(self.y)
        cm.volatility = EWMAVariance()
        cm.fit(update_freq=0, disp=DISPLAY)
        cm.volatility = RiskMetrics2006()
        cm.fit(update_freq=0, disp=DISPLAY)

        ar = ARX(self.y, lags=5)
        ar.volatility = EWMAVariance()
        ar.fit(update_freq=0, disp=DISPLAY)
        ar.volatility = RiskMetrics2006()
        ar.fit(update_freq=0, disp=DISPLAY)
        assert "tau0" in str(ar.volatility)
        assert "tau1" in str(ar.volatility)
        assert "kmax" in str(ar.volatility)

    def test_egarch(self):
        cm = ConstantMean(self.y)
        cm.volatility = EGARCH()
        res = cm.fit(update_freq=0, disp=DISPLAY)
        summ = res.summary()
        assert "Df Model:                            1" in str(summ)
        cm.distribution = StudentsT()
        cm.fit(update_freq=0, disp=DISPLAY)

    def test_multiple_lags(self):
        """Smoke test to ensure models estimate with multiple lags"""
        vp = {"garch": GARCH, "egarch": EGARCH, "harch": HARCH, "arch": ARCH}
        cm = ConstantMean(self.y)
        for name, process in vp.items():
            cm.volatility = process()
            cm.fit(update_freq=0, disp=DISPLAY)
            for p in [1, 2, 3]:
                for o in [1, 2, 3]:
                    for q in [1, 2, 3]:
                        if name in ("arch",):
                            cm.volatility = process(p=p + o + q)
                            cm.fit(update_freq=0, disp=DISPLAY)
                        elif name in ("harch",):
                            cm.volatility = process(lags=[p, p + o, p + o + q])
                            cm.fit(update_freq=0, disp=DISPLAY)
                        else:
                            cm.volatility = process(p=p, o=o, q=q)
                            cm.fit(update_freq=0, disp=DISPLAY)

    def test_first_last_obs(self):
        ar = ARX(self.y, lags=5, hold_back=100)
        res = ar.fit(update_freq=0, disp=DISPLAY)
        resids = res.resid
        resid_copy = resids.copy()
        resid_copy[:100] = np.nan
        assert_equal(resids, resid_copy)

        ar.volatility = GARCH()
        res = ar.fit(update_freq=0, disp=DISPLAY)
        resids = res.resid
        resid_copy = resids.copy()
        resid_copy[:100] = np.nan
        assert_equal(resids, resid_copy)

        ar = ARX(self.y, lags=5)
        ar.volatility = GARCH()
        res = ar.fit(update_freq=0, last_obs=500, disp=DISPLAY)
        resids = res.resid
        resid_copy = resids.copy()
        resid_copy[500:] = np.nan
        assert_equal(resids, resid_copy)

        ar = ARX(self.y, lags=5, hold_back=100)
        ar.volatility = GARCH()
        res = ar.fit(update_freq=0, last_obs=500, disp=DISPLAY)
        resids = res.resid
        resid_copy = resids.copy()
        resid_copy[:100] = np.nan
        resid_copy[500:] = np.nan
        assert_equal(resids, resid_copy)

        vol = res.conditional_volatility
        vol_copy = vol.copy()
        vol_copy[:100] = np.nan
        vol_copy[500:] = np.nan
        assert_equal(vol, vol_copy)
        assert_equal(self.y.shape[0], vol.shape[0])

        ar = ARX(self.y, lags=5)
        ar.volatility = GARCH()
        res = ar.fit(update_freq=0, last_obs=500, disp=DISPLAY)
        resids = res.resid
        resid_copy = resids.copy()
        resid_copy[:5] = np.nan
        resid_copy[500:] = np.nan
        assert_equal(resids, resid_copy)

    def test_date_first_last_obs(self):
        y = self.y_series

        cm = ConstantMean(y)
        res = cm.fit(last_obs=y.index[900], disp=DISPLAY)

        cm = ConstantMean(y)
        res2 = cm.fit(last_obs=900, disp=DISPLAY)

        assert_equal(res.resid.values, res2.resid.values)

    def test_align(self):
        dates = pd.date_range("2000-01-01", "2010-01-01", freq="M")
        columns = ["h." + "{0:>02}".format(h + 1) for h in range(10)]
        forecasts = pd.DataFrame(self.rng.randn(120, 10), index=dates, columns=columns)

        aligned = _align_forecast(forecasts.copy(), align="origin")
        assert_frame_equal(aligned, forecasts)

        aligned = _align_forecast(forecasts.copy(), align="target")
        direct = forecasts.copy()
        for i in range(10):
            direct.iloc[(i + 1) :, i] = direct.iloc[: (120 - i - 1), i].values
            direct.iloc[: (i + 1), i] = np.nan
        assert_frame_equal(aligned, direct)

        with pytest.raises(ValueError):
            _align_forecast(forecasts, align="unknown")

    def test_fixed_user_parameters(self):
        am = arch_model(self.y_series)
        res = am.fit(disp=DISPLAY)
        fixed_res = am.fix(res.params)
        assert_series_equal(
            res.conditional_volatility, fixed_res.conditional_volatility
        )
        assert_series_equal(res.params, fixed_res.params)
        assert_equal(res.aic, fixed_res.aic)
        assert_equal(res.bic, fixed_res.bic)
        assert_equal(res.loglikelihood, fixed_res.loglikelihood)
        assert_equal(res.num_params, fixed_res.num_params)
        # Smoke for summary
        fixed_res.summary()

    def test_fixed_user_parameters_new_model(self):
        am = arch_model(self.y_series)
        res = am.fit(disp=DISPLAY)
        new_am = arch_model(self.y_series)
        fixed_res = new_am.fix(res.params)
        assert_series_equal(
            res.conditional_volatility, fixed_res.conditional_volatility
        )
        assert_series_equal(res.params, fixed_res.params)
        assert_equal(res.aic, fixed_res.aic)
        assert_equal(res.bic, fixed_res.bic)
        assert_equal(res.loglikelihood, fixed_res.loglikelihood)
        assert_equal(res.num_params, fixed_res.num_params)

        # Test first and last dates
        am = arch_model(self.y_series)
        res = am.fit(disp=DISPLAY, first_obs=100, last_obs=900)
        new_am = arch_model(self.y_series)
        fixed_res = new_am.fix(res.params, first_obs=100, last_obs=900)
        assert_series_equal(res.params, fixed_res.params)
        assert_equal(res.aic, fixed_res.aic)
        assert_equal(res.bic, fixed_res.bic)
        assert_equal(res.loglikelihood, fixed_res.loglikelihood)
        assert_equal(res.num_params, fixed_res.num_params)

    def test_output_options(self):
        am = arch_model(self.y_series)
        orig_stdout = sys.stdout

        try:
            sio = StringIO()
            sys.stdout = sio
            am.fit(disp=DISPLAY)
            sio.seek(0)
            output = sio.read()
            assert len(output) == 0
        finally:
            sys.stdout = orig_stdout

    def test_convergence_warning(self):
        y = np.array(
            [
                0.83277114,
                0.45194014,
                -0.33475561,
                -0.49463896,
                0.54715787,
                1.11895382,
                1.31280266,
                0.81464021,
                0.8532107,
                1.0967188,
                0.9346354,
                0.92289249,
                1.01339085,
                1.071065,
                1.42413486,
                1.15392453,
                1.10929691,
                0.96162061,
                0.96489515,
                0.93250153,
                1.34509807,
                1.80951607,
                1.66313783,
                1.38610821,
                1.26381761,
            ]
        )
        am = arch_model(y, mean="ARX", lags=10, p=5, q=0)

        warning = ConvergenceWarning if SP_LT_14 else None
        with pytest.warns(warning):
            am.fit(disp=DISPLAY)

        with pytest.warns(warning):
            am.fit(show_warning=True, disp=DISPLAY)

        with pytest.warns(DataScaleWarning):
            am.fit(show_warning=False, disp=DISPLAY)

    def test_first_after_last(self):
        am = arch_model(self.y_series)
        with pytest.raises(ValueError):
            am.fit(disp=DISPLAY, first_obs=500, last_obs=480)

        with pytest.raises(ValueError):
            am.fit(
                disp=DISPLAY,
                first_obs=self.y_series.index[500],
                last_obs=self.y_series.index[480],
            )

    def test_sample_adjustment(self):
        am = arch_model(self.y_series, vol="Constant")
        res = am.fit(disp=DISPLAY)

        res_adj = am.fit(disp=DISPLAY, first_obs=0, last_obs=self.y_series.shape[0] + 1)
        assert_equal(res.resid.values, res_adj.resid.values)
        assert_equal(res.params.values, res_adj.params.values)

        res = am.fit(disp=DISPLAY, first_obs=100)
        assert res.fit_start == 100
        res_adj = am.fit(disp=DISPLAY, first_obs=self.y_series.index[100])
        assert_equal(res.params.values, res_adj.params.values)
        assert_equal(res.resid.values, res_adj.resid.values)

        res = am.fit(disp=DISPLAY, last_obs=900)
        res2 = am.fit(disp=DISPLAY, last_obs=self.y_series.index[900])
        assert_equal(res.params.values, res2.params.values)
        assert_equal(res.resid.values, res2.resid.values)

        res = am.fit(disp=DISPLAY, first_obs=100, last_obs=900)
        res2 = am.fit(
            disp=DISPLAY,
            first_obs=self.y_series.index[100],
            last_obs=self.y_series.index[900],
        )
        assert_equal(res.params.values, res2.params.values)
        assert_equal(res.resid.values, res2.resid.values)

    def test_model_obs_equivalence(self):
        """Tests models that should use the same observation"""
        am = arch_model(self.y_series.iloc[100:900])
        res = am.fit(disp=DISPLAY)
        am = arch_model(self.y_series)
        res2 = am.fit(disp=DISPLAY, first_obs=100, last_obs=900)
        index = self.y_series.index
        res3 = am.fit(disp=DISPLAY, first_obs=index[100], last_obs=index[900])
        assert_equal(res.params.values, res2.params.values)
        assert_equal(res2.params.values, res3.params.values)

        am = arch_model(self.y_series, hold_back=100)
        res4 = am.fit(disp=DISPLAY, last_obs=900)
        assert_equal(res.params.values, res4.params.values)

    def test_model_obs_equivalence_ar(self):
        """Tests models that should use the same observation"""
        am = arch_model(self.y_series.iloc[100:900], mean="AR", lags=[1, 2, 4])
        res = am.fit(disp=DISPLAY)
        am = arch_model(self.y_series, mean="AR", lags=[1, 2, 4])
        res2 = am.fit(disp=DISPLAY, first_obs=100, last_obs=900)
        index = self.y_series.index
        res3 = am.fit(disp=DISPLAY, first_obs=index[100], last_obs=index[900])
        assert_almost_equal(res.params.values, res2.params.values, decimal=4)
        assert_almost_equal(res2.params.values, res3.params.values, decimal=4)

        am = arch_model(self.y_series, mean="AR", lags=[1, 2, 4], hold_back=100)
        res4 = am.fit(disp=DISPLAY, first_obs=4, last_obs=900)
        assert_almost_equal(res.params.values, res4.params.values, decimal=4)
        assert am.hold_back == 100

    def test_constant_mean_fixed_variance(self):
        rng = RandomState(1234)
        variance = 2 + rng.standard_normal(self.y.shape[0]) ** 2.0
        std = np.sqrt(variance)
        y = pd.Series(
            std * rng.standard_normal(self.y_series.shape[0]), index=self.y_series.index
        )

        mod = ConstantMean(y, volatility=FixedVariance(variance))
        res = mod.fit(disp=DISPLAY)
        res.summary()
        assert len(res.params) == 2
        assert "scale" in res.params.index

        mod = ARX(self.y_series, lags=[1, 2, 3], volatility=FixedVariance(variance))
        res = mod.fit(disp=DISPLAY)
        assert len(res.params) == 5
        assert "scale" in res.params.index

        mod = ARX(
            self.y_series,
            lags=[1, 2, 3],
            volatility=FixedVariance(variance, unit_scale=True),
        )
        res = mod.fit(disp=DISPLAY)
        assert len(res.params) == 4
        assert "scale" not in res.params.index

    def test_optimization_options(self):
        norm = Normal(random_state=RandomState([12891298, 843084]))
        am = arch_model(None)
        am.distribution = norm
        data = am.simulate(np.array([0.0, 0.1, 0.1, 0.85]), 2500)
        am = arch_model(data.data)
        std = am.fit(disp=DISPLAY)
        loose = am.fit(tol=1e-2, disp=DISPLAY)
        assert std.loglikelihood >= loose.loglikelihood
        with warnings.catch_warnings(record=True) as w:
            short = am.fit(options={"maxiter": 3}, disp=DISPLAY)
        assert len(w) == 1
        assert std.loglikelihood >= short.loglikelihood
        assert short.convergence_flag != 0

    def test_little_or_no_data(self):
        mod = HARX(self.y[:24], lags=[1, 5, 22])
        with pytest.raises(ValueError):
            mod.fit(disp=DISPLAY)
        mod = HARX(None, lags=[1, 5, 22])
        with pytest.raises(RuntimeError):
            mod.fit(disp=DISPLAY)

    def test_empty_mean(self):
        mod = HARX(
            self.y,
            None,
            None,
            False,
            volatility=ConstantVariance(),
            distribution=Normal(),
        )
        res = mod.fit(disp=DISPLAY)

        mod = ZeroMean(self.y, volatility=ConstantVariance(), distribution=Normal())
        res_z = mod.fit(disp=DISPLAY)

        assert res.num_params == res_z.num_params
        assert_series_equal(res.params, res_z.params)
        assert res.loglikelihood == res_z.loglikelihood


@pytest.mark.parametrize(
    "volatility",
    [GARCH, EGARCH, RiskMetrics2006, EWMAVariance, HARCH, ConstantVariance],
)
def test_backcast(volatility, simulated_data):
    zm = ZeroMean(simulated_data, volatility=volatility())
    res = zm.fit(disp=DISPLAY)
    bc = zm.volatility.backcast(np.asarray(res.resid))
    if volatility is EGARCH:
        bc = np.exp(bc)
    res2 = zm.fit(backcast=bc, disp=DISPLAY)
    assert_array_almost_equal(res.params, res2.params)
    if volatility is RiskMetrics2006:
        zm.fit(backcast=bc[0], disp=DISPLAY)


def test_backcast_error(simulated_data):
    zm = ZeroMean(simulated_data, volatility=GARCH())
    with pytest.raises(ValueError):
        zm.fit(backcast=-1, disp=DISPLAY)
    zm = ZeroMean(simulated_data, volatility=RiskMetrics2006())
    with pytest.raises(ValueError):
        zm.fit(backcast=np.ones(100), disp=DISPLAY)


@pytest.mark.parametrize(
    "volatility",
    [
        ConstantVariance,
        GARCH,
        EGARCH,
        FIGARCH,
        HARCH,
        MIDASHyperbolic,
        RiskMetrics2006,
        EWMAVariance,
    ],
)
def test_fit_smoke(simulated_data, volatility):
    zm = ZeroMean(simulated_data, volatility=volatility())
    zm.fit(disp=DISPLAY)


def test_arch_lm(simulated_data):
    zm = ZeroMean(simulated_data, volatility=GARCH())
    res = zm.fit(disp=DISPLAY)
    wald = res.arch_lm_test()
    nobs = simulated_data.shape[0]
    df = int(np.ceil(12.0 * np.power(nobs / 100.0, 1 / 4.0)))
    assert wald.df == df
    assert "Standardized" not in wald.null
    assert "Standardized" not in wald.alternative
    assert "H0: Standardized" not in wald.__repr__()
    assert "heteroskedastic" in wald.__repr__()

    resids2 = pd.Series(res.resid ** 2)
    data = [resids2.shift(i) for i in range(df + 1)]
    data = pd.concat(data, 1).dropna()
    lhs = data.iloc[:, 0]
    rhs = smtools.add_constant(data.iloc[:, 1:])
    ols_res = smlm.OLS(lhs, rhs).fit()
    assert_almost_equal(wald.stat, nobs * ols_res.rsquared)
    assert len(wald.critical_values) == 3
    assert "10%" in wald.critical_values

    wald = res.arch_lm_test(lags=5)
    assert wald.df == 5
    assert_almost_equal(wald.pval, 1 - stats.chi2(5).cdf(wald.stat))

    wald = res.arch_lm_test(standardized=True)
    assert wald.df == df
    assert "Standardized" in wald.null
    assert "Standardized" in wald.alternative
    assert_almost_equal(wald.pval, 1 - stats.chi2(df).cdf(wald.stat))
    assert "H0: Standardized" in wald.__repr__()


def test_autoscale():
    rs = np.random.RandomState(34254321)
    dist = Normal(random_state=rs)
    am = arch_model(None)
    am.distribution = dist
    data = am.simulate([0, 0.0001, 0.05, 0.94], nobs=1000)
    am = arch_model(data.data)
    with pytest.warns(DataScaleWarning):
        res = am.fit(disp=DISPLAY)
    assert_almost_equal(res.scale, 1.0)

    am = arch_model(data.data, rescale=True)
    res_auto = am.fit(disp=DISPLAY)
    assert_almost_equal(res_auto.scale, 10.0)

    am = arch_model(10 * data.data)
    res_manual = am.fit(disp=DISPLAY)
    assert_series_equal(res_auto.params, res_manual.params)

    res_no = arch_model(data.data, rescale=False).fit(disp=DISPLAY)
    assert res_no.scale == 1.0

    am = arch_model(10000 * data.data, rescale=True)
    res_big = am.fit(disp=DISPLAY)
    assert_almost_equal(res_big.scale, 0.1)


def test_no_variance():
    mod = arch_model(np.ones(100))
    with pytest.warns(ConvergenceWarning):
        mod.fit(disp=DISPLAY)


def test_1d_exog():
    rs = np.random.RandomState(329302)
    y = rs.standard_normal((300))
    x = rs.standard_normal((300))
    am = arch_model(y, x, mean="ARX", lags=2, vol="ARCH", q=0)
    res = am.fit()
    am = arch_model(y, x[:, None], mean="ARX", lags=2, vol="ARCH", q=0)
    res2 = am.fit()
    assert_series_equal(res.params, res2.params)


def test_harx_lag_spec(simulated_data):
    harx_1 = HARX(simulated_data, lags=[1, 5, 22])
    harx_2 = HARX(simulated_data, lags=[1, 5, 22], use_rotated=True)
    harx_3 = HARX(simulated_data, lags=[[1, 1, 1], [1, 5, 22]])
    harx_4 = HARX(simulated_data, lags=[[1, 2, 6], [1, 5, 22]])

    r2 = harx_1.fit().rsquared
    assert_almost_equal(harx_2.fit().rsquared, r2)
    assert_almost_equal(harx_3.fit().rsquared, r2)
    assert_almost_equal(harx_4.fit().rsquared, r2)


def test_backcast_restricted(simulated_data):
    # GH 440
    mod = arch_model(simulated_data)
    res = mod.fit(disp="off")
    subset = (
        simulated_data[100:600]
        if isinstance(simulated_data, np.ndarray)
        else simulated_data.iloc[100:600]
    )
    mod_restricted = arch_model(subset)
    res_restricted = mod_restricted.fit(disp="off")
    res_limited = mod.fit(first_obs=100, last_obs=600, disp="off")
    assert_almost_equal(res_restricted.model._backcast, res_restricted.model._backcast)
    assert np.abs(res.model._backcast - res_limited.model._backcast) > 1e-8



from typing import Any, Union

from numpy import recarray
from pandas import DataFrame


def dataset_loader(dataset: Any) -> Union[recarray, DataFrame]:
    """Load a dataset using the new syntax is possible"""
    try:
        return dataset.load(as_pandas=True).data
    except TypeError:
        return dataset.load().data



from arch.compat.numba import PerformanceWarning

import numpy as np
import pytest

from arch.univariate.recursions_python import arch_recursion

try:
    import numba  # noqa: F401

    HAS_NUMBA = True
except ImportError:
    HAS_NUMBA = False


@pytest.mark.skipif(HAS_NUMBA, reason="Can only test when numba is not available.")
def test_performance_warning():
    parameters = np.array([1, 0.1])
    nobs = 100
    resids = np.ones(nobs)
    sigma2 = resids.copy()
    p = 1
    backcast = 1.0
    var_bounds = np.empty((nobs, 2))
    var_bounds[:, 0] = 0.0
    var_bounds[:, 1] = 1.0e14
    with pytest.warns(PerformanceWarning):
        arch_recursion(parameters, resids, sigma2, p, nobs, backcast, var_bounds)



"""
Utility functions that do not explicitly relate to Volatility modeling
"""
from arch.compat.pandas import is_datetime64_any_dtype

from abc import ABCMeta
import datetime as dt
from typing import Any, Dict, Hashable, List, Optional, Sequence, Tuple, Type, Union

import numpy as np
from pandas import DataFrame, DatetimeIndex, Index, NaT, Series, Timestamp, to_datetime

from arch.typing import AnyPandas, ArrayLike, DateLike, NDArray
from arch.vendor import cached_property

__all__ = [
    "ensure1d",
    "parse_dataframe",
    "DocStringInheritor",
    "date_to_index",
    "cutoff_to_index",
    "ensure2d",
    "AbstractDocStringInheritor",
    "find_index",
]

deprecation_doc: str = """
{func} has been moved.  Please use {new_location}.{func}.
"""


def ensure1d(
    x: Union[int, float, Sequence[Union[int, float]], ArrayLike],  # noqa: E231
    name: Optional[Hashable],
    series: bool = False,
) -> Union[NDArray, Series]:
    if isinstance(x, Series):
        if not isinstance(x.name, str):
            x.name = str(x.name)
        if series:
            return x
        else:
            return np.asarray(x)

    if isinstance(x, DataFrame):
        if x.shape[1] != 1:
            raise ValueError(f"{name} must be squeezable to 1 dimension")
        if not series:
            return np.asarray(x.iloc[:, 0])
        x_series = Series(x.iloc[:, 0], x.index)
        if not isinstance(x_series.name, str):
            x_series.name = str(x_series.name)
        return x_series

    x_arr = np.squeeze(np.asarray(x))
    if x_arr.ndim == 0:
        x_arr = x_arr[None]
    elif x_arr.ndim != 1:
        raise ValueError(f"{name} must be squeezable to 1 dimension")

    if series:
        return Series(x_arr, name=name)
    else:
        return x_arr


def ensure2d(
    x: Union[
        Sequence[Union[float, int]], Sequence[Sequence[Union[float, int]]], ArrayLike
    ],
    name: str,
) -> Union[DataFrame, NDArray]:
    if isinstance(x, Series):
        return DataFrame(x)
    elif isinstance(x, DataFrame):
        return x
    elif isinstance(x, np.ndarray):
        if x.ndim == 0:
            return np.asarray([[x]])
        elif x.ndim == 1:
            return x[:, None]
        elif x.ndim == 2:
            return x
        else:
            raise ValueError("Variable " + name + "must be 2d or reshapable to 2d")
    else:
        raise TypeError("Variable " + name + "must be a Series, DataFrame or ndarray.")


def parse_dataframe(
    x: Optional[ArrayLike], name: Union[str, List[str]]
) -> Union[
    Tuple[Index, Index],
    Tuple[List[Optional[Hashable]], Index],
    Tuple[List[str], NDArray],
]:
    if x is None:
        return [name], np.empty(0)
    if isinstance(x, DataFrame):
        return x.columns, x.index
    elif isinstance(x, Series):
        return [x.name], x.index
    else:
        if not isinstance(name, list):
            name = [name]
        return name, np.arange(np.squeeze(x).shape[0])


class DocStringInheritor(type):
    """
    A variation on
    http://groups.google.com/group/comp.lang.python/msg/26f7b4fcb4d66c95
    by Paul McGuire
    """

    def __new__(
        mcs, name: str, bases: Tuple[Type, ...], clsdict: Dict[str, Any]
    ) -> Any:
        if not ("__doc__" in clsdict and clsdict["__doc__"]):
            for mro_cls in (mro_cls for base in bases for mro_cls in base.mro()):
                doc = mro_cls.__doc__
                if doc:
                    clsdict["__doc__"] = doc
                    break
        for attr, attribute in clsdict.items():
            if not attribute.__doc__:
                for mro_cls in (
                    mro_cls
                    for base in bases
                    for mro_cls in base.mro()
                    if hasattr(mro_cls, attr)
                ):
                    doc = getattr(getattr(mro_cls, attr), "__doc__")
                    if doc:
                        if isinstance(attribute, cached_property):
                            attribute.func.__doc__ = doc
                            clsdict[attr] = cached_property(attribute.func)
                        elif isinstance(attribute, property):
                            clsdict[attr] = property(
                                attribute.__get__,
                                attribute.__set__,
                                attribute.__delete__,
                                doc,
                            )
                        else:
                            attribute.__doc__ = doc
                        break
        return type.__new__(mcs, name, bases, clsdict)


class ConcreteClassMeta(ABCMeta):
    def __init__(cls, *args: Any, **kwargs: Any):
        super(ConcreteClassMeta, cls).__init__(*args, **kwargs)
        missing: List[str] = getattr(cls, "__abstractmethods__", [])
        if missing:
            missing_meth = ", ".join(missing)
            raise TypeError(
                f"{cls.__name__} has not implemented abstract methods {missing_meth}"
            )


class AbstractDocStringInheritor(ConcreteClassMeta, DocStringInheritor):
    pass


def date_to_index(
    date: Union[str, dt.date, dt.datetime, np.datetime64, Timestamp],
    date_index: Union[DatetimeIndex, NDArray],
) -> int:
    """
    Looks up a date in an array of dates

    Parameters
    ----------
    date : string, datetime or datetime64
        Date to use when returning the index
    date_index : ndarray
        Index data containing datetime64 values

    Returns
    -------
    int
        Index location

    Notes
    -----
    Assumes dates are increasing and unique.
    """
    if not is_datetime64_any_dtype(date_index):
        raise ValueError("date_index must be a datetime64 array")
    if isinstance(date_index, DatetimeIndex):
        if not np.all((np.diff(date_index.values).astype(dtype=np.int64)) > 0):
            raise ValueError("date_index is not monotonic and unique")
    if not isinstance(date, (dt.datetime, np.datetime64, str)):
        raise ValueError("date must be a datetime, datetime64 or string")
    elif isinstance(date, Timestamp):
        assert isinstance(date_index, DatetimeIndex)
        if date_index.tzinfo is not None:
            date = date.tz_convert("GMT").tz_localize(None)
        date = date.to_datetime64()
    elif isinstance(date, dt.datetime):
        date = np.datetime64(date)
    elif isinstance(date, str):
        orig_date = date
        try:
            date = np.datetime64(to_datetime(date, errors="coerce"))
        except (ValueError, TypeError):
            raise ValueError("date:" + orig_date + " cannot be parsed to a date.")

    if isinstance(date_index, DatetimeIndex):
        if date_index.tzinfo is not None:
            date_index = date_index.tz_convert("GMT").tz_localize(None)

    date_index = np.asarray(date_index)

    locs = np.nonzero(date_index <= date)[0]
    if locs.shape[0] == 0:
        return 0

    loc = locs.max()
    in_array = np.any(date_index == date)
    if not in_array:
        loc += 1

    return int(loc)


def cutoff_to_index(
    cutoff: Optional[Union[int, DateLike]], index: DatetimeIndex, default: int
) -> int:
    """
    Converts a cutoff to a numerical index

    Parameters
    ----------
    cutoff : {None, str, datetime, datetime64, Timestamp)
        The cutoff point to use
    index : DatetimeIndex
        Pandas index
    default : int
        The value to return if cutoff is None

    Returns
    -------
    int
        Integer value where
    """
    int_index = default
    if isinstance(cutoff, (str, dt.datetime, np.datetime64, Timestamp)):
        int_index = date_to_index(cutoff, index)
    elif isinstance(cutoff, int) or issubclass(cutoff.__class__, np.integer):
        assert cutoff is not None
        int_index = int(cutoff)

    return int_index


def find_index(s: AnyPandas, index: Union[int, DateLike]) -> int:
    """
    Returns the numeric index for a string or datetime

    Parameters
    ----------
    s : Series or DataFrame
        Series or DataFrame to use in lookup
    index : datetime-like, str
        Index value, either a string convertible to a datetime or a datetime

    Returns
    -------
    int
        Integer location of index value
    """
    if isinstance(index, (int, np.int64)):
        return int(index)
    date_index = to_datetime(index, errors="coerce")

    if date_index is NaT:
        raise ValueError(f"{index} cannot be converted to datetime")
    loc = np.argwhere(s.index == date_index).squeeze()
    if loc.size == 0:
        raise ValueError("index not found")
    return int(loc)



import datetime as dt

import numpy as np
from numpy.random import RandomState
from numpy.testing import assert_allclose
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

from arch.tests.univariate.test_variance_forecasting import preserved_state
from arch.univariate import HARX, arch_model
from arch.univariate.mean import _ar_forecast, _ar_to_impulse


class TestForecasting(object):
    @classmethod
    def setup_class(cls):
        cls.rng = RandomState(12345)
        am = arch_model(None, mean="Constant", vol="Constant")
        data = am.simulate(np.array([0.0, 10.0]), 1000)
        data.index = pd.date_range("2000-01-01", periods=data.index.shape[0])
        cls.zero_mean = data.data

        am = arch_model(None, mean="AR", vol="Constant", lags=[1])
        data = am.simulate(np.array([1.0, 0.9, 2]), 1000)
        data.index = pd.date_range("2000-01-01", periods=data.index.shape[0])
        cls.ar1 = data.data

        am = arch_model(None, mean="AR", vol="Constant", lags=[1, 2])
        data = am.simulate(np.array([1.0, 1.9, -0.95, 2]), 1000)
        data.index = pd.date_range("2000-01-01", periods=data.index.shape[0])
        cls.ar2 = data.data

        am = arch_model(None, mean="HAR", vol="Constant", lags=[1, 5, 22])
        data = am.simulate(np.array([1.0, 0.4, 0.3, 0.2, 2]), 1000)
        data.index = pd.date_range("2000-01-01", periods=data.index.shape[0])
        cls.har3 = data.data

        am = arch_model(None, mean="AR", vol="GARCH", lags=[1, 2], p=1, q=1)
        data = am.simulate(np.array([1.0, 1.9, -0.95, 0.05, 0.1, 0.88]), 1000)
        data.index = pd.date_range("2000-01-01", periods=data.index.shape[0])
        cls.ar2_garch = data.data

    def test_ar_forecasting(self):
        params = np.array([0.9])
        forecasts = _ar_forecast(self.zero_mean, 5, 0, 0.0, params)
        expected = np.zeros((1000, 5))
        expected[:, 0] = 0.9 * self.zero_mean.values
        for i in range(1, 5):
            expected[:, i] = 0.9 * expected[:, i - 1]
        assert_allclose(forecasts, expected)

        params = np.array([0.5, -0.3, 0.2])
        forecasts = _ar_forecast(self.zero_mean, 5, 2, 0.0, params)
        expected = np.zeros((998, 8))
        expected[:, 0] = self.zero_mean.iloc[0:-2]
        expected[:, 1] = self.zero_mean.iloc[1:-1]
        expected[:, 2] = self.zero_mean.iloc[2:]
        for i in range(3, 8):
            expected[:, i] = (
                0.5 * expected[:, i - 1]
                - 0.3 * expected[:, i - 2]
                + 0.2 * expected[:, i - 3]
            )
        fill = np.empty((2, 5))
        fill.fill(np.nan)
        expected = np.concatenate((fill, expected[:, 3:]))
        assert_allclose(forecasts, expected)

    def test_ar_to_impulse(self):
        arp = np.array([0.9])
        impulses = _ar_to_impulse(20, arp)
        expected = 0.9 ** np.arange(20)
        assert_allclose(impulses, expected)

        arp = np.array([0.5, 0.3])
        impulses = _ar_to_impulse(20, arp)
        comp = np.array([arp, [1, 0]])
        a = comp.copy()
        expected = np.ones(20)
        for i in range(1, 20):
            expected[i] = a[0, 0]
            a = a.dot(comp)
        assert_allclose(impulses, expected)

        arp = np.array([1.5, 0.0, -0.7])
        impulses = _ar_to_impulse(20, arp)
        comp = np.array([arp, [1, 0, 0], [0, 1, 0]])
        a = comp.copy()
        expected = np.ones(20)
        for i in range(1, 20):
            expected[i] = a[0, 0]
            a = a.dot(comp)
        assert_allclose(impulses, expected)

    def test_zero_mean_forecast(self):
        am = arch_model(self.zero_mean, mean="Zero", vol="Constant")
        res = am.fit()
        fcast = res.forecast(res.params, horizon=3)
        alt_fcast = res.forecast(horizon=3)
        assert_frame_equal(fcast.mean, alt_fcast.mean)
        assert_frame_equal(fcast.variance, alt_fcast.variance)
        assert_frame_equal(fcast.residual_variance, alt_fcast.residual_variance)

        assert np.all(np.asarray(np.isnan(fcast.mean[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.variance[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.residual_variance[:-1])))

        params = np.asarray(res.params)
        assert np.all(0.0 == fcast.mean.iloc[-1])
        assert_allclose(fcast.variance.iloc[-1], np.ones(3) * params[0])
        assert_allclose(fcast.residual_variance.iloc[-1], np.ones(3) * params[0])

        res = am.fit(last_obs=500)
        params = np.asarray(res.params)
        fcast = res.forecast(horizon=3)
        assert np.all(np.asarray(np.isnan(fcast.mean[:499])))
        assert np.all(np.asarray(np.isnan(fcast.variance[:499])))
        assert np.all(np.asarray(np.isnan(fcast.residual_variance[:499])))

        assert np.all(np.asarray(0.0 == fcast.mean[499:]))
        assert_allclose(fcast.variance.iloc[499:], np.ones((501, 3)) * params[0])
        assert_allclose(
            fcast.residual_variance.iloc[499:], np.ones((501, 3)) * params[0]
        )
        with pytest.raises(ValueError, match="horizon must be an integer >= 1"):
            res.forecast(horizon=0)

    def test_frame_labels(self):
        am = arch_model(self.zero_mean, mean="Zero", vol="Constant")
        res = am.fit()
        fcast = res.forecast(horizon=12)
        assert fcast.mean.shape[1] == 12
        assert fcast.variance.shape[1] == 12
        assert fcast.residual_variance.shape[1] == 12
        for i in range(1, 13):
            if i < 10:
                col = "h.0" + str(i)
            else:
                col = "h." + str(i)

            assert col in fcast.mean.columns
            assert col in fcast.variance.columns
            assert col in fcast.residual_variance.columns

    def test_ar1_forecast(self):
        am = arch_model(self.ar1, mean="AR", vol="Constant", lags=[1])
        res = am.fit()

        fcast = res.forecast(horizon=5, start=0)
        params = np.asarray(res.params)
        direct = self.ar1.values

        for i in range(5):
            direct = params[0] + params[1] * direct
            assert_allclose(direct, fcast.mean.iloc[:, i])
            scale = np.sum((params[1] ** np.arange(i + 1)) ** 2.0)
            var = fcast.variance.iloc[1:, i]
            assert_allclose(var, scale * params[2] * np.ones_like(var))

        assert np.all(np.asarray(fcast.residual_variance[1:] == params[2]))

        fcast = res.forecast(horizon=5)
        params = np.asarray(res.params)
        assert np.all(np.asarray(np.isnan(fcast.mean[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.variance[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.residual_variance[:-1])))
        assert np.all(np.asarray(fcast.residual_variance.iloc[-1] == params[-1]))
        means = np.zeros(5)
        means[0] = params[0] + params[1] * self.ar1.iloc[-1]
        for i in range(1, 5):
            means[i] = params[0] + params[1] * means[i - 1]
        assert_allclose(means, fcast.mean.iloc[-1].values)

    def test_constant_mean_forecast(self):
        am = arch_model(self.zero_mean, mean="Constant", vol="Constant")
        res = am.fit()
        fcast = res.forecast(horizon=5)

        assert np.all(np.asarray(np.isnan(fcast.mean[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.variance[:-1])))
        assert np.all(np.asarray(np.isnan(fcast.residual_variance[:-1])))
        params = np.asarray(res.params)
        assert_allclose(params[0] * np.ones(5), fcast.mean.iloc[-1])
        assert_allclose(params[1] * np.ones(5), fcast.variance.iloc[-1])
        assert_allclose(params[1] * np.ones(5), fcast.residual_variance.iloc[-1])

        assert fcast.mean.shape == (self.zero_mean.shape[0], 5)
        assert fcast.variance.shape == (self.zero_mean.shape[0], 5)
        assert fcast.residual_variance.shape == (self.zero_mean.shape[0], 5)

    def test_ar2_forecast(self):
        am = arch_model(self.ar2, mean="AR", vol="Constant", lags=[1, 2])
        res = am.fit()

        fcast = res.forecast(horizon=5)
        params = np.asarray(res.params)
        expected = np.zeros(7)
        expected[:2] = self.ar2.iloc[-2:]
        for i in range(2, 7):
            expected[i] = (
                params[0] + params[1] * expected[i - 1] + params[2] * expected[i - 2]
            )

        expected = expected[2:]
        assert np.all(np.asarray(np.isnan(fcast.mean.iloc[:-1])))
        assert_allclose(fcast.mean.iloc[-1], expected)

        expected = np.zeros(5)
        comp = np.array([res.params.iloc[1:3], [1, 0]])
        a = np.eye(2)
        for i in range(5):
            expected[i] = a[0, 0]
            a = a.dot(comp)
        expected = res.params.iloc[-1] * np.cumsum(expected ** 2)
        assert_allclose(fcast.variance.iloc[-1], expected)

        expected = np.empty((1000, 5))
        expected[:2] = np.nan
        expected[2:] = res.params.iloc[-1]

        fcast = res.forecast(horizon=5, start=1)
        expected = np.zeros((999, 7))
        expected[:, 0] = self.ar2.iloc[0:-1]
        expected[:, 1] = self.ar2.iloc[1:]
        for i in range(2, 7):
            expected[:, i] = (
                params[0]
                + params[1] * expected[:, i - 1]
                + params[2] * expected[:, i - 2]
            )
        fill = np.empty((1, 5))
        fill.fill(np.nan)
        expected = np.concatenate((fill, expected[:, 2:]))
        assert_allclose(np.asarray(fcast.mean), expected)

        expected = np.empty((1000, 5))
        expected[:2] = np.nan
        expected[2:] = res.params.iloc[-1]
        assert_allclose(np.asarray(fcast.residual_variance), expected)

        with pytest.raises(ValueError):
            res.forecast(horizon=5, start=0)

    def test_har_forecast(self):
        am = arch_model(self.har3, mean="HAR", vol="Constant", lags=[1, 5, 22])
        res = am.fit()
        fcast_1 = res.forecast(horizon=1)
        fcast_5 = res.forecast(horizon=5)
        assert_allclose(fcast_1.mean, fcast_5.mean.iloc[:, :1])

        with pytest.raises(ValueError):
            res.forecast(horizon=1, start=0)
        with pytest.raises(ValueError):
            res.forecast(horizon=1, start=20)

        fcast_66 = res.forecast(horizon=66, start=21)
        expected = np.empty((1000, 66 + 22))
        expected.fill(np.nan)
        for i in range(22):
            if i < 21:
                expected[21:, i] = self.har3.iloc[i : (-21 + i)]
            else:
                expected[21:, i] = self.har3.iloc[i:]
        params = np.asarray(res.params)
        const = params[0]
        arp = np.zeros(22)
        arp[0] = params[1]
        arp[:5] += params[2] / 5
        arp[:22] += params[3] / 22
        arp_rev = arp[::-1]
        for i in range(22, 88):
            expected[:, i] = const + expected[:, i - 22 : i].dot(arp_rev)
        expected = expected[:, 22:]
        assert_allclose(fcast_66.mean, expected)

        expected[:22] = np.nan
        expected[22:] = res.params.iloc[-1]
        assert_allclose(fcast_66.residual_variance, expected)

        impulse = _ar_to_impulse(66, arp)
        expected = expected * np.cumsum(impulse ** 2)
        assert_allclose(fcast_66.variance, expected)

    def test_forecast_start_alternatives(self):
        am = arch_model(self.har3, mean="HAR", vol="Constant", lags=[1, 5, 22])
        res = am.fit()
        date = self.har3.index[21]
        fcast_1 = res.forecast(start=21)
        fcast_2 = res.forecast(start=date)
        for field in ("mean", "variance", "residual_variance"):
            assert_frame_equal(getattr(fcast_1, field), getattr(fcast_2, field))
        pydt = dt.datetime(date.year, date.month, date.day)
        fcast_2 = res.forecast(start=pydt)
        for field in ("mean", "variance", "residual_variance"):
            assert_frame_equal(getattr(fcast_1, field), getattr(fcast_2, field))

        strdt = pydt.strftime("%Y-%m-%d")
        fcast_2 = res.forecast(start=strdt)
        for field in ("mean", "variance", "residual_variance"):
            assert_frame_equal(getattr(fcast_1, field), getattr(fcast_2, field))

        npydt = np.datetime64(pydt).astype("M8[ns]")
        fcast_2 = res.forecast(start=npydt)
        for field in ("mean", "variance", "residual_variance"):
            assert_frame_equal(getattr(fcast_1, field), getattr(fcast_2, field))

        with pytest.raises(ValueError):
            date = self.har3.index[20]
            res.forecast(start=date)

        with pytest.raises(ValueError):
            date = self.har3.index[0]
            res.forecast(start=date)

        fcast_0 = res.forecast()
        fcast_1 = res.forecast(start=999)
        fcast_2 = res.forecast(start=self.har3.index[999])
        for field in ("mean", "variance", "residual_variance"):
            assert_frame_equal(getattr(fcast_0, field), getattr(fcast_1, field))

            assert_frame_equal(getattr(fcast_0, field), getattr(fcast_2, field))

    def test_fit_options(self):
        am = arch_model(self.zero_mean, mean="Constant", vol="Constant")
        res = am.fit(first_obs=100)
        res.forecast()
        res = am.fit(last_obs=900)
        res.forecast()
        res = am.fit(first_obs=100, last_obs=900)
        res.forecast()
        res.forecast(start=100)
        res.forecast(start=200)
        am = arch_model(self.zero_mean, mean="Constant", vol="Constant", hold_back=20)
        res = am.fit(first_obs=100)
        res.forecast()

    def test_ar1_forecast_simulation(self):
        am = arch_model(self.ar1, mean="AR", vol="GARCH", lags=[1])
        res = am.fit(disp="off")

        with preserved_state(self.rng):
            forecast = res.forecast(horizon=5, start=0, method="simulation")

        y = np.asarray(self.ar1)
        index = self.ar1.index
        t = y.shape[0]
        params = np.array(res.params)
        resids = np.asarray(y[1:] - params[0] - params[1] * y[:-1])
        vol = am.volatility
        params = np.array(res.params)
        backcast = vol.backcast(resids)
        var_bounds = vol.variance_bounds(resids)
        rng = am.distribution.simulate([])
        vfcast = vol.forecast(
            params[2:],
            resids,
            backcast,
            var_bounds,
            start=0,
            method="simulation",
            rng=rng,
            horizon=5,
        )
        const, ar = params[0], params[1]
        means = np.zeros((t, 5))
        means[:, 0] = const + ar * y
        for i in range(1, 5):
            means[:, i] = const + ar * means[:, i - 1]
        means = pd.DataFrame(
            means, index=index, columns=["h.{0}".format(j) for j in range(1, 6)]
        )
        assert_frame_equal(means, forecast.mean)
        var = np.concatenate([[[np.nan] * 5], vfcast.forecasts])
        rv = pd.DataFrame(
            var, index=index, columns=["h.{0}".format(j) for j in range(1, 6)]
        )
        assert_frame_equal(rv, forecast.residual_variance)

        lrv = rv.copy()
        for i in range(5):
            weights = (ar ** np.arange(i + 1)) ** 2
            weights = weights[:, None]
            lrv.iloc[:, i : i + 1] = rv.values[:, : i + 1].dot(weights[::-1])
        assert_frame_equal(lrv, forecast.variance)

    def test_ar1_forecast_bootstrap(self):
        am = arch_model(self.ar1, mean="AR", vol="GARCH", lags=[1])
        res = am.fit(disp="off")
        rs = np.random.RandomState(98765432)
        state = rs.get_state()
        forecast = res.forecast(
            horizon=5, start=900, method="bootstrap", random_state=rs
        )
        rs.set_state(state)
        repeat = res.forecast(horizon=5, start=900, method="bootstrap", random_state=rs)
        assert_frame_equal(forecast.mean, repeat.mean)
        assert_frame_equal(forecast.variance, repeat.variance)

    def test_ar2_garch11(self):
        pass

    def test_first_obs(self):
        y = self.ar2_garch
        mod = arch_model(y)
        res = mod.fit(disp="off", first_obs=y.index[100])
        mod = arch_model(y[100:])
        res2 = mod.fit(disp="off")
        assert_allclose(res.params, res2.params)
        mod = arch_model(y)
        res3 = mod.fit(disp="off", first_obs=100)
        assert res.fit_start == 100
        assert_allclose(res.params, res3.params)

        forecast = res.forecast(horizon=3)
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[:-1])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[-1])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[:-1])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[-1])))

        forecast = res.forecast(horizon=3, start=y.index[100])
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[100:])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[100:])))

        forecast = res.forecast(horizon=3, start=100)
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[100:])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[100:])))

        with pytest.raises(ValueError):
            res.forecast(horizon=3, start=y.index[98])

        res = mod.fit(disp="off")
        forecast = res.forecast(horizon=3)
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[:-1])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[-1])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[:-1])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[-1])))

        forecast = res.forecast(horizon=3, start=y.index[100])
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[100:])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[:100])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[100:])))
        forecast = res.forecast(horizon=3, start=0)
        assert np.all(np.asarray(np.isfinite(forecast.mean)))
        assert np.all(np.asarray(np.isfinite(forecast.variance)))

        mod = arch_model(y, mean="AR", lags=[1, 2])
        res = mod.fit(disp="off")
        with pytest.raises(ValueError):
            res.forecast(horizon=3, start=0)

        forecast = res.forecast(horizon=3, start=1)
        assert np.all(np.asarray(np.isnan(forecast.mean.iloc[0])))
        assert np.all(np.asarray(np.isfinite(forecast.mean.iloc[1:])))
        assert np.all(np.asarray(np.isnan(forecast.variance.iloc[0])))
        assert np.all(np.asarray(np.isfinite(forecast.variance.iloc[2:])))

    def test_last_obs(self):
        y = self.ar2_garch
        mod = arch_model(y)
        res = mod.fit(disp="off", last_obs=y.index[900])
        res_2 = mod.fit(disp="off", last_obs=900)
        assert_allclose(res.params, res_2.params)
        mod = arch_model(y[:900])
        res_3 = mod.fit(disp="off")
        assert_allclose(res.params, res_3.params)

    def test_first_last_obs(self):
        y = self.ar2_garch
        mod = arch_model(y)
        res = mod.fit(disp="off", first_obs=y.index[100], last_obs=y.index[900])
        res_2 = mod.fit(disp="off", first_obs=100, last_obs=900)
        assert_allclose(res.params, res_2.params)
        mod = arch_model(y.iloc[100:900])
        res_3 = mod.fit(disp="off")
        assert_allclose(res.params, res_3.params)

        mod = arch_model(y)
        res_4 = mod.fit(disp="off", first_obs=100, last_obs=y.index[900])
        assert_allclose(res.params, res_4.params)

    def test_holdback_first_obs(self):
        y = self.ar2_garch
        mod = arch_model(y, hold_back=20)
        res_holdback = mod.fit(disp="off")
        mod = arch_model(y)
        res_first_obs = mod.fit(disp="off", first_obs=20)
        assert_allclose(res_holdback.params, res_first_obs.params)

        with pytest.raises(ValueError):
            res_holdback.forecast(start=18)

    def test_holdback_lastobs(self):
        y = self.ar2_garch
        mod = arch_model(y, hold_back=20)
        res_holdback_last_obs = mod.fit(disp="off", last_obs=800)
        mod = arch_model(y)
        res_first_obs_last_obs = mod.fit(disp="off", first_obs=20, last_obs=800)
        assert_allclose(res_holdback_last_obs.params, res_first_obs_last_obs.params)
        mod = arch_model(y[20:800])
        res_direct = mod.fit(disp="off")
        assert_allclose(res_direct.params, res_first_obs_last_obs.params)

        with pytest.raises(ValueError):
            res_holdback_last_obs.forecast(start=18)

    def test_holdback_ar(self):
        y = self.ar2_garch
        mod = arch_model(y, mean="AR", lags=1, hold_back=1)
        res_holdback = mod.fit(disp="off")
        mod = arch_model(y, mean="AR", lags=1)
        res = mod.fit(disp="off")
        assert_allclose(res_holdback.params, res.params, rtol=1e-4, atol=1e-4)

    def test_forecast_exogenous_regressors(self):
        y = 10 * self.rng.randn(1000, 1)
        x = self.rng.randn(1000, 2)
        am = HARX(y=y, x=x, lags=[1, 2])
        res = am.fit()
        fcasts = res.forecast(horizon=1, start=1)

        const, har01, har02, ex0, ex1, _ = res.params
        y_01 = y[1:-1]
        y_02 = (y[1:-1] + y[0:-2]) / 2
        x0 = x[2:, :1]
        x1 = x[2:, 1:2]
        direct = const + har01 * y_01 + har02 * y_02 + ex0 * x0 + ex1 * x1
        direct = np.vstack(([[np.nan]], direct, [[np.nan]]))
        direct = pd.DataFrame(direct, columns=["h.1"])
        assert_allclose(np.asarray(direct), fcasts.mean)

        fcasts2 = res.forecast(horizon=2, start=1)
        assert fcasts2.mean.shape == (1000, 2)
        assert fcasts2.mean.isnull().all()["h.2"]
        assert_frame_equal(fcasts.mean, fcasts2.mean[["h.1"]])



import glob
import os
import sys

import pytest

SKIP = True

try:
    import jupyter_client

    # matplotlib is required for most notebooks
    import matplotlib  # noqa: F401
    from nbconvert.preprocessors import ExecutePreprocessor
    import nbformat

    kernels = jupyter_client.kernelspec.find_kernel_specs()
    SKIP = False
except ImportError:  # pragma: no cover
    pytestmark = pytest.mark.skip(reason="Required packages not available")

SLOW_NOTEBOOKS = ["multiple-comparison_examples.ipynb"]
if bool(os.environ.get("ARCH_TEST_SLOW_NOTEBOOKS", False)):
    SLOW_NOTEBOOKS = []
kernel_name = "python%s" % sys.version_info.major

head, _ = os.path.split(__file__)
NOTEBOOK_DIR = os.path.abspath(os.path.join(head, "..", "..", "examples"))

nbs = sorted(glob.glob(os.path.join(NOTEBOOK_DIR, "*.ipynb")))
ids = list(map(lambda s: os.path.split(s)[-1].split(".")[0], nbs))
if not nbs:  # pragma: no cover
    pytest.mark.skip(reason="No notebooks found and so no tests run")


@pytest.fixture(params=nbs, ids=ids)
def notebook(request):
    return request.param


@pytest.mark.slow
@pytest.mark.skipif(SKIP, reason="Required packages not available")
def test_notebook(notebook):
    nb_name = os.path.split(notebook)[-1]
    if nb_name in SLOW_NOTEBOOKS:
        pytest.skip("Notebook is too slow to test")
    nb = nbformat.read(notebook, as_version=4)
    ep = ExecutePreprocessor(allow_errors=False, timeout=240, kernel_name=kernel_name)
    ep.preprocess(nb, {"metadata": {"path": NOTEBOOK_DIR}})



import numpy as np
from numpy.random import RandomState
from numpy.testing import assert_array_almost_equal, assert_array_equal, assert_equal
import pandas as pd
from pandas.testing import assert_frame_equal
import pytest

from arch.utility.timeseries import ColumnNameConflict, add_trend


@pytest.fixture(scope="function")
def rng():
    return RandomState(12345)


def test_add_trend_err():
    with pytest.raises(ValueError, match="One and only one"):
        add_trend(x=None, trend="ctt", nobs=None)


def test_add_trend_prepend(rng):
    n = 10
    x = rng.randn(n, 1)
    trend_1 = add_trend(x, trend="ct", prepend=True)
    trend_2 = add_trend(x, trend="ct", prepend=False)
    assert_equal(trend_1[:, :2], trend_2[:, 1:])


def test_add_time_trend_dataframe(rng):
    n = 10
    x = rng.randn(n, 1)
    x = pd.DataFrame(x, columns=["col1"])
    trend_1 = add_trend(x, trend="t")
    assert_array_almost_equal(np.asarray(trend_1["trend"]), np.arange(1.0, n + 1))


def test_add_trend_prepend_dataframe(rng):
    n = 10
    x = rng.randn(n, 1)
    x = pd.DataFrame(x, columns=["col1"])
    trend_1 = add_trend(x, trend="ct", prepend=True)
    trend_2 = add_trend(x, trend="ct", prepend=False)
    assert_frame_equal(trend_1.iloc[:, :2], trend_2.iloc[:, 1:])


def test_add_trend_duplicate_name():
    x = pd.DataFrame(np.zeros((10, 1)), columns=["trend"])
    with pytest.warns(ColumnNameConflict):
        add_trend(x, trend="ct")
        y = add_trend(x, trend="ct")

    assert "const" in y.columns
    assert "trend_0" in y.columns


def test_add_trend_c():
    x = np.zeros((10, 1))
    y = add_trend(x, trend="c")
    assert np.all(y[:, 1] == 1.0)


def test_add_trend_ct():
    n = 20
    x = np.zeros((20, 1))
    y = add_trend(x, trend="ct")
    assert np.all(y[:, 1] == 1.0)
    assert_equal(y[0, 2], 1.0)
    assert_array_almost_equal(np.diff(y[:, 2]), np.ones((n - 1)))


def test_add_trend_ctt():
    n = 10
    x = np.zeros((n, 1))
    y = add_trend(x, trend="ctt")
    assert np.all(y[:, 1] == 1.0)
    assert y[0, 2] == 1.0
    assert_array_almost_equal(np.diff(y[:, 2]), np.ones((n - 1)))
    assert y[0, 3] == 1.0
    assert_array_almost_equal(np.diff(y[:, 3]), np.arange(3.0, 2.0 * n, 2.0))


def test_add_trend_t():
    n = 20
    x = np.zeros((20, 1))
    y = add_trend(x, trend="t")
    assert y[0, 1] == 1.0
    assert_array_almost_equal(np.diff(y[:, 1]), np.ones((n - 1)))


def test_add_trend_no_input():
    n = 100
    y = add_trend(x=None, trend="ct", nobs=n)
    assert np.all(y[:, 0] == 1.0)
    assert y[0, 1] == 1.0
    assert_array_almost_equal(np.diff(y[:, 1]), np.ones((n - 1)))


def test_skip_constant():
    x = np.ones((100, 1))
    appended = add_trend(x, trend="c", has_constant="add")
    assert_array_equal(np.ones((100, 2)), appended)
    appended = add_trend(x, trend="c", has_constant="skip")
    assert_array_equal(np.ones((100, 1)), appended)


def test_errors():
    n = 100
    with pytest.raises(ValueError):
        add_trend(x=None, trend="unknown", nobs=n)
    with pytest.raises(ValueError):
        add_trend(x=None, trend="ct")
    x = np.ones((100, 1))
    with pytest.raises(ValueError):
        add_trend(x, trend="ct", has_constant="raise")


def test_trend_n_nobs():
    assert add_trend(nobs=100, trend="n").shape == (100, 0)
    assert add_trend(np.empty((100, 2)), trend="n").shape == (100, 2)



import datetime as dt

import numpy as np
from numpy.random import RandomState
from numpy.testing import assert_equal
from pandas import DataFrame, Series, Timedelta, date_range
import pytest

from arch import doc
from arch.univariate.base import implicit_constant
from arch.utility.array import (
    DocStringInheritor,
    cutoff_to_index,
    date_to_index,
    ensure1d,
    ensure2d,
    find_index,
    parse_dataframe,
)


class TestUtils(object):
    @classmethod
    def setup_class(cls):
        cls.rng = RandomState(12345)

    def test_ensure1d(self):
        out = ensure1d(1.0, "y")
        assert_equal(out, np.array([1.0]))
        out = ensure1d(np.arange(5.0), "y")
        assert_equal(out, np.arange(5.0))
        out = ensure1d(np.arange(5.0)[:, None], "y")
        assert_equal(out, np.arange(5.0))
        in_array = np.reshape(np.arange(16.0), (4, 4))
        with pytest.raises(ValueError):
            ensure1d(in_array, "y")

        y = Series(np.arange(5.0))
        ys = ensure1d(y, "y")
        assert isinstance(ys, np.ndarray)
        ys = ensure1d(y, "y", True)
        assert isinstance(ys, Series)
        y = DataFrame(y)
        ys = ensure1d(y, "y")
        assert isinstance(ys, np.ndarray)
        ys = ensure1d(y, "y", True)
        assert isinstance(ys, Series)
        y.columns = [1]
        ys = ensure1d(y, "y", True)
        assert isinstance(ys, Series)
        assert ys.name == "1"
        y = Series(np.arange(5.0), name="series")
        ys = ensure1d(y, "y")
        assert isinstance(ys, np.ndarray)
        ys = ensure1d(y, "y", True)
        assert isinstance(ys, Series)
        y = DataFrame(y)
        ys = ensure1d(y, "y")
        assert isinstance(ys, np.ndarray)
        ys = ensure1d(y, "y", True)
        assert isinstance(ys, Series)
        ys.name = 1
        ys = ensure1d(ys, None, True)
        assert isinstance(ys, Series)
        assert ys.name == "1"
        y = DataFrame(np.reshape(np.arange(10), (5, 2)))
        with pytest.raises(ValueError):
            ensure1d(y, "y")

    def test_ensure2d(self):
        s = Series([1, 2, 3], name="x")
        df = ensure2d(s, "x")
        assert isinstance(df, DataFrame)

        df2 = ensure2d(df, "x")
        assert df is df2

        npa = ensure2d(s.values, "x")
        assert isinstance(npa, np.ndarray)
        assert npa.ndim == 2

        npa = ensure2d(np.array(1.0), "x")
        assert isinstance(npa, np.ndarray)
        assert npa.ndim == 2

        with pytest.raises(ValueError):
            ensure2d(np.array([[[1]]]), "x")
        with pytest.raises(TypeError):
            ensure2d([1], "x")

    def test_parse_dataframe(self):
        s = Series(np.arange(10.0), name="variable")
        out = parse_dataframe(s, "y")
        assert_equal(out[1], np.arange(10.0))
        assert_equal(out[0], ["variable"])
        df = DataFrame(s)
        out = parse_dataframe(df, "y")
        assert_equal(out[1], np.arange(10.0))
        assert_equal(out[0], ["variable"])
        out = parse_dataframe(np.arange(10.0), "y")
        assert_equal(out[1], np.arange(10.0))
        assert_equal(out[0], ["y"])

        out = parse_dataframe(None, "name")
        assert out[0] == ["name"]
        assert isinstance(out[1], np.ndarray)
        assert out[1].shape == (0,)

    def test_implicit_constant(self):
        x = self.rng.standard_normal((1000, 2))
        assert not implicit_constant(x)
        x[:, 0] = 1.0
        assert implicit_constant(x)
        x = self.rng.standard_normal((1000, 3))
        x[:, 0] = x[:, 0] > 0
        x[:, 1] = 1 - x[:, 0]
        assert implicit_constant(x)

    def test_docstring_inheritor(self):
        class A(object, metaclass=DocStringInheritor):
            """
            Docstring
            """

        class B(A):
            pass

        assert_equal(B.__doc__, A.__doc__)

    def test_date_to_index(self):
        dr = date_range("20000101", periods=3000, freq="W")
        y = Series(np.arange(3000.0), index=dr)
        date_index = y.index

        index = date_to_index(date_index[0], date_index)
        assert_equal(index, 0)
        index = date_to_index(date_index[-1], date_index)
        assert_equal(index, date_index.shape[0] - 1)

        index = date_to_index("2009-08-02", date_index)
        assert_equal(index, 500)
        index = date_to_index("2009-08-04", date_index)
        assert_equal(index, 501)
        index = date_to_index("2009-08-01", date_index)
        assert_equal(index, 500)
        index = date_to_index(dt.datetime(2009, 8, 1), date_index)
        assert_equal(index, 500)
        with pytest.raises(ValueError):
            date_to_index(dt.date(2009, 8, 1), date_index)
        z = y + 0.0
        z.index = np.arange(3000)
        num_index = z.index
        with pytest.raises(ValueError):
            date_to_index(dt.datetime(2009, 8, 1), num_index)
        idx = date_range("1999-12-31", periods=3)

        df = DataFrame([1, 2, 3], index=idx[::-1])
        with pytest.raises(ValueError):
            date_to_index(idx[0], df.index)

        df = DataFrame([1, 2, 3], index=[idx[0]] * 3)
        with pytest.raises(ValueError):
            date_to_index(idx[0], df.index)

        with pytest.raises(ValueError):
            date_to_index("NaT", idx)

        #  check whether this also works for a localized datetimeindex
        date_index = date_range("20000101", periods=3000, freq="W", tz="Europe/Berlin")
        index = date_to_index(date_index[0], date_index)
        assert_equal(index, 0)

    def test_date_to_index_timestamp(self):
        dr = date_range("20000101", periods=3000, freq="W")
        y = Series(np.arange(3000.0), index=dr)
        date_index = y.index
        date = y.index[1000]
        date_pydt = date.to_pydatetime()
        date_npdt = date.to_datetime64()
        date_str = date_pydt.strftime("%Y-%m-%d")

        index = date_to_index(date, date_index)
        index_pydt = date_to_index(date_pydt, date_index)
        index_npdt = date_to_index(date_npdt, date_index)
        index_str = date_to_index(date_str, date_index)
        assert_equal(index, 1000)
        assert_equal(index, index_npdt)
        assert_equal(index, index_pydt)
        assert_equal(index, index_str)

    def test_(self):
        dr = date_range("20000101", periods=3000, freq="W")
        y = Series(np.arange(3000.0), index=dr)
        date_index = y.index

        date = date_index[1000] + Timedelta(1, "D")
        date_pydt = date.to_pydatetime()
        date_npdt = date.to_datetime64()
        date_str = date_pydt.strftime("%Y-%m-%d")
        index = date_to_index(date, date_index)
        index_pydt = date_to_index(date_pydt, date_index)
        index_npdt = date_to_index(date_npdt, date_index)
        index_str = date_to_index(date_str, date_index)
        assert_equal(index, 1001)
        assert_equal(index, index_npdt)
        assert_equal(index, index_pydt)
        assert_equal(index, index_str)

        date = date_index[0] - Timedelta(1, "D")
        index = date_to_index(date, date_index)
        assert_equal(index, 0)

        date_pydt = date.to_pydatetime()
        date_npdt = date.to_datetime64()
        date_str = date_pydt.strftime("%Y-%m-%d")
        index_pydt = date_to_index(date_pydt, date_index)
        index_npdt = date_to_index(date_npdt, date_index)
        index_str = date_to_index(date_str, date_index)
        assert_equal(index, index_npdt)
        assert_equal(index, index_pydt)
        assert_equal(index, index_str)

    def test_cutoff_to_index(self):
        dr = date_range("20000101", periods=3000, freq="W")
        y = Series(np.arange(3000.0), index=dr)
        date_index = y.index
        assert cutoff_to_index(1000, date_index, 0) == 1000
        assert cutoff_to_index(int(1000), date_index, 0) == 1000
        assert cutoff_to_index(np.int16(1000), date_index, 0) == 1000
        assert cutoff_to_index(np.int64(1000), date_index, 0) == 1000
        assert cutoff_to_index(date_index[1000], date_index, 0) == 1000
        assert cutoff_to_index(None, date_index, 1000) == 1000

    def test_find_index(self):
        index = date_range("2000-01-01", periods=5000)
        series = Series(np.arange(len(index)), index=index, name="test")
        df = DataFrame(series)
        assert_equal(find_index(series, "2000-01-01"), 0)
        assert_equal(find_index(series, series.index[0]), 0)
        assert_equal(find_index(series, series.index[3000]), 3000)
        assert_equal(find_index(series, series.index[3000].to_pydatetime()), 3000)
        npy_date = np.datetime64(series.index[3000].to_pydatetime())
        found_loc = find_index(series, npy_date)
        assert_equal(found_loc, 3000)
        with pytest.raises(ValueError):
            find_index(series, "bad-date")
        with pytest.raises(ValueError):
            find_index(series, "1900-01-01")

        assert_equal(find_index(df, "2000-01-01"), 0)
        assert_equal(find_index(df, df.index[0]), 0)
        assert_equal(find_index(df, df.index[3000]), 3000)
        assert_equal(find_index(df, df.index[3000].to_pydatetime()), 3000)
        found_loc = find_index(df, np.datetime64(df.index[3000].to_pydatetime()))
        assert_equal(found_loc, 3000)
        with pytest.raises(ValueError):
            find_index(df, "bad-date")
        with pytest.raises(ValueError):
            find_index(df, "1900-01-01")

        idx = find_index(df, 1)
        assert idx == 1


def test_doc():
    doc()



from pandas import DataFrame

from arch.data.utility import load_file


def load() -> DataFrame:

    """
    Load the Fama-French factor data used in the examples

    Returns
    -------
    data : DataFrame
        Data set containing excess market, size and value factors and the
        risk-free rate

    Notes
    -----
    Provided by Ken French,
    http://mba.tuck.dartmouth.edu/pages/faculty/ken.french/data_library.html
    """
    return load_file(__file__, "frenchdata.csv.gz")



from arch.compat.statsmodels import dataset_loader

from numpy import diff, log
from numpy.random import RandomState
from numpy.testing import assert_almost_equal
import pytest
from statsmodels.datasets import macrodata

from arch.utility import cov_nw


class TestVarNW(object):
    @classmethod
    def setup_class(cls):

        cls.rng = RandomState(12345)
        cls.cpi = log(dataset_loader(macrodata)["cpi"])
        cls.inflation = diff(cls.cpi)

    def test_cov_nw(self):
        y = self.inflation
        simple_cov = cov_nw(y, lags=0)
        e = y - y.mean()
        assert_almost_equal(e.dot(e) / e.shape[0], simple_cov)

    def test_cov_nw_ddof(self):
        y = self.inflation
        simple_cov = cov_nw(y, lags=0, ddof=1)
        e = y - y.mean()
        n = e.shape[0]
        assert_almost_equal(e.dot(e) / (n - 1), simple_cov)

    def test_cov_nw_no_demean(self):
        y = self.inflation
        simple_cov = cov_nw(y, lags=0, demean=False)
        assert_almost_equal(y.dot(y) / y.shape[0], simple_cov)

    def test_cov_nw_2d(self):
        y = self.rng.randn(100, 2)
        simple_cov = cov_nw(y, lags=0)
        e = y - y.mean(0)
        assert_almost_equal(e.T.dot(e) / e.shape[0], simple_cov)

    def test_cov_nw_2d_2lags(self):
        y = self.rng.randn(100, 2)
        e = y - y.mean(0)
        gamma_0 = e.T.dot(e)
        gamma_1 = e[1:].T.dot(e[:-1])
        gamma_2 = e[2:].T.dot(e[:-2])
        w1, w2 = 1.0 - (1.0 / 3.0), 1.0 - (2.0 / 3.0)
        expected = (
            gamma_0 + w1 * (gamma_1 + gamma_1.T) + w2 * (gamma_2 + gamma_2.T)
        ) / 100.0
        assert_almost_equal(cov_nw(y, lags=2), expected)

    def test_cov_nw_axis(self):
        y = self.rng.randn(100, 2)
        e = y - y.mean(0)
        gamma_0 = e.T.dot(e)
        gamma_1 = e[1:].T.dot(e[:-1])
        gamma_2 = e[2:].T.dot(e[:-2])
        w1, w2 = 1.0 - (1.0 / 3.0), 1.0 - (2.0 / 3.0)
        expected = (
            gamma_0 + w1 * (gamma_1 + gamma_1.T) + w2 * (gamma_2 + gamma_2.T)
        ) / 100.0
        assert_almost_equal(cov_nw(y.T, lags=2, axis=1), expected)

    def test_errors(self):
        y = self.rng.randn(100, 2)
        with pytest.raises(ValueError):
            cov_nw(y, 200)
        with pytest.raises(ValueError):
            cov_nw(y, axis=3)
        with pytest.raises(ValueError):
            cov_nw(y, ddof=200)



from typing import Tuple

import numpy as np
from numpy.testing import assert_allclose
import pytest
from statsmodels.iolib.summary import Summary

from arch.typing import NDArray
from arch.unitroot._phillips_ouliaris import (
    CriticalValueWarning,
    phillips_ouliaris_cv,
    phillips_ouliaris_pval,
)
from arch.unitroot.cointegration import phillips_ouliaris
from arch.utility.timeseries import add_trend


def z_tests(z: NDArray, lag: int, trend: str) -> Tuple[float, float]:
    z = add_trend(z, trend=trend)
    u = z
    if z.shape[1] > 1:
        delta = np.linalg.lstsq(z[:, 1:], z[:, 0], rcond=None)[0]
        u = z[:, 0] - z[:, 1:] @ delta
    alpha = (u[:-1].T @ u[1:]) / (u[:-1].T @ u[:-1])
    k = u[1:] - alpha * u[:-1]
    nobs = u.shape[0]
    one_sided_strict = 0.0
    for i in range(1, lag + 1):
        w = 1 - i / (lag + 1)
        one_sided_strict += 1 / nobs * w * k[i:].T @ k[:-i]
    u2 = u[:-1].T @ u[:-1]
    z = (alpha - 1) - nobs * one_sided_strict / u2
    z_a = nobs * z
    long_run = k.T @ k / nobs + 2 * one_sided_strict
    se = np.sqrt(long_run / u2)
    z_t = z / se
    print(u2, long_run, one_sided_strict, z)
    return float(z_a), float(z_t)


def p_tests(z: NDArray, lag: int, trend: str) -> Tuple[float, float]:
    x, y = z[:, 1:], z[:, 0]
    nobs = x.shape[0]
    x = add_trend(x, trend=trend)
    beta = np.linalg.lstsq(x, y, rcond=None)[0]
    u = y - x @ beta
    z_lead = z[1:]
    z_lag = add_trend(z[:-1], trend=trend)
    phi = np.linalg.lstsq(z_lag, z_lead, rcond=None)[0]
    xi = z_lead - z_lag @ phi

    omega = xi.T @ xi / nobs
    for i in range(1, lag + 1):
        w = 1 - i / (lag + 1)
        gamma = xi[i:].T @ xi[:-i] / nobs
        omega += w * (gamma + gamma.T)
    omega21 = omega[0, 1:]
    omega22 = omega[1:, 1:]
    omega112 = omega[0, 0] - np.squeeze(omega21.T @ np.linalg.inv(omega22) @ omega21)
    denom = u.T @ u / nobs
    p_u = nobs * omega112 / denom

    tr = add_trend(nobs=z.shape[0], trend=trend)
    if tr.shape[1]:
        z = z - tr @ np.linalg.lstsq(tr, z, rcond=None)[0]
    else:
        z = z - z[0]
    m_zz = z.T @ z / nobs
    p_z = nobs * (omega @ np.linalg.inv(m_zz)).trace()
    return p_u, p_z


@pytest.mark.parametrize("trend", ["n", "c", "ct", "ctt"])
@pytest.mark.parametrize("test_type", ["Za", "Zt", "Pu", "Pz"])
@pytest.mark.parametrize("kernel", ["bartlett", "parzen", "quadratic-spectral"])
@pytest.mark.parametrize("bandwidth", [None, 10])
@pytest.mark.parametrize("force_int", [True, False])
def test_smoke(trivariate_data, trend, test_type, kernel, bandwidth, force_int):
    y, x = trivariate_data
    res = phillips_ouliaris(
        y,
        x,
        trend=trend,
        test_type=test_type,
        kernel=kernel,
        bandwidth=bandwidth,
        force_int=force_int,
    )
    assert isinstance(res.stat, float)


def test_errors(trivariate_data):
    y, x = trivariate_data
    with pytest.raises(ValueError, match="kernel is not a known estimator."):
        phillips_ouliaris(y, x, kernel="fancy-kernel")
    with pytest.raises(ValueError, match="Unknown test_type: z-alpha."):
        phillips_ouliaris(y, x, test_type="z-alpha")


@pytest.mark.parametrize("trend", ["n", "c", "ct", "ctt"])
@pytest.mark.parametrize("bandwidth", [0, 1, 10])
def test_z_test_direct(trivariate_data, trend, bandwidth):
    y, x = trivariate_data
    y = np.asarray(y)
    x = np.asarray(x)
    z = np.column_stack([y, x])
    ref_z_a, ref_z_t = z_tests(z, bandwidth, trend)
    z_a = phillips_ouliaris(y, x, trend=trend, bandwidth=bandwidth, test_type="Za")
    z_t = phillips_ouliaris(y, x, trend=trend, bandwidth=bandwidth, test_type="Zt")
    assert_allclose(z_a.stat, ref_z_a)
    assert_allclose(z_t.stat, ref_z_t)
    assert isinstance(z_a.summary(), Summary)
    assert isinstance(z_t.summary(), Summary)


@pytest.mark.parametrize("trend", ["n", "c", "ct", "ctt"])
@pytest.mark.parametrize("bandwidth", [0, 1, 10])
def test_p_test_direct(trivariate_data, trend, bandwidth):
    y, x = trivariate_data
    y = np.asarray(y)
    x = np.asarray(x)
    z = np.column_stack([y, x])
    ref_p_u, ref_p_z = p_tests(z, bandwidth, trend)
    p_u = phillips_ouliaris(y, x, trend=trend, bandwidth=bandwidth, test_type="Pu")
    p_z = phillips_ouliaris(y, x, trend=trend, bandwidth=bandwidth, test_type="Pz")
    assert_allclose(p_u.stat, ref_p_u)
    assert_allclose(p_z.stat, ref_p_z)
    assert isinstance(p_u.summary(), Summary)
    assert isinstance(p_z.summary(), Summary)


def test_cv_exceptions():
    with pytest.raises(ValueError, match="test_type must be one of"):
        phillips_ouliaris_cv("unknown", "c", 2, 500)
    with pytest.raises(ValueError, match="trend must by one of:"):
        phillips_ouliaris_cv("Pu", "unknown", 2, 500)
    with pytest.raises(ValueError, match="The number of stochastic trends must"):
        phillips_ouliaris_cv("Pu", "ct", 25, 500)
    with pytest.warns(CriticalValueWarning):
        phillips_ouliaris_cv("Pu", "ct", 2, 10)


def test_pval_exceptions():
    with pytest.raises(ValueError, match="test_type must be one of"):
        phillips_ouliaris_pval(3.0, "unknown", "c", 2)
    with pytest.raises(ValueError, match="trend must by one of:"):
        phillips_ouliaris_pval(3.0, "Pu", "unknown", 2)
    with pytest.raises(ValueError, match="The number of stochastic trends must"):
        phillips_ouliaris_pval(3.0, "Pu", "ct", 25)


def test_pval_extremes():
    assert phillips_ouliaris_pval(3.0, "Zt", "n", 2) == 1.0
    assert phillips_ouliaris_pval(-3000.0, "Zt", "n", 2) == 0.0
    # Above and below tau-star
    above = phillips_ouliaris_pval(1.0, "Zt", "n", 2)
    below = phillips_ouliaris_pval(0.0, "Zt", "n", 2)
    assert above > below


def test_auto_bandwidth(trivariate_data):
    y, x = trivariate_data
    res = phillips_ouliaris(y, x)
    assert isinstance(res.summary(), Summary)
    assert int(res.bandwidth) != res.bandwidth
    res = phillips_ouliaris(y, x, force_int=True)
    assert int(res.bandwidth) == res.bandwidth



#!python
#cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True

import numpy as np

cimport numpy as np
from libc.float cimport DBL_MAX
from libc.math cimport exp, fabs, log, sqrt

__all__ = ['harch_recursion', 'arch_recursion', 'garch_recursion', 'egarch_recursion',
           'midas_recursion', 'figarch_recursion', 'figarch_weights', 'aparch_recursion']

cdef double LNSIGMA_MAX = log(DBL_MAX)

np.import_array()

cdef inline void bounds_check(double* sigma2, double* var_bounds):
    if sigma2[0] < var_bounds[0]:
        sigma2[0] = var_bounds[0]
    elif sigma2[0] > var_bounds[1]:
        if sigma2[0] > DBL_MAX:
            sigma2[0] = var_bounds[1] + 1000
        else:
            sigma2[0] = var_bounds[1] + log(sigma2[0] / var_bounds[1])


def harch_recursion(double[::1] parameters,
                    double[::1] resids,
                    double[::1] sigma2,
                    int[::1] lags,
                    int nobs,
                    double backcast,
                    double[:, ::1] var_bounds):
    """
    Parameters
    ----------
    parameters : 1-d array, float64
        Model parameters
    resids : 1-d array, float64
        Residuals to use in the recursion
    sigma2 : 1-d array, float64
        Conditional variances with same shape as resids
    lags : 1-d array, int
        Lag lengths in the HARCH
    nobs : int
        Length of resids
    backcast : float64
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        variances for each time period
    """
    cdef Py_ssize_t t, i, num_lags
    cdef int j
    cdef double param
    num_lags = lags.shape[0]

    for t in range(nobs):
        sigma2[t] = parameters[0]
        for i in range(num_lags):
            param = parameters[i + 1] / lags[i]
            for j in range(lags[i]):
                if (t - j - 1) >= 0:
                    sigma2[t] += param * resids[t - j - 1] * resids[t - j - 1]
                else:
                    sigma2[t] += param * backcast
        bounds_check(&sigma2[t], &var_bounds[t, 0])

    return np.asarray(sigma2)

def arch_recursion(double[::1] parameters,
                   double[::1] resids,
                   double[::1] sigma2,
                   int p,
                   int nobs,
                   double backcast,
                   double[:, ::1] var_bounds):
    """
    Parameters
    ----------
    parameters : 1-d array, float64
        Model parameters
    resids : 1-d array, float64
        Residuals to use in the recursion
    sigma2 : 1-d array, float64
        Conditional variances with same shape as resids
    p : int
        Number of lags in ARCH model
    nobs : int
        Length of resids
    backcast : float64
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        variances for each time period
    """

    cdef Py_ssize_t t
    cdef int i
    cdef double param

    for t in range(nobs):
        sigma2[t] = parameters[0]
        for i in range(p):
            if (t - i - 1) < 0:
                sigma2[t] += parameters[i + 1] * backcast
            else:
                sigma2[t] += parameters[i + 1] * resids[t - i - 1] * \
                             resids[t - i - 1]
        bounds_check(&sigma2[t], &var_bounds[t, 0])

    return np.asarray(sigma2)

def garch_recursion(double[::1] parameters,
                    double[::1] fresids,
                    double[::1] sresids,
                    double[::1] sigma2,
                    int p,
                    int o,
                    int q,
                    int nobs,
                    double backcast,
                    double[:, ::1] var_bounds):
    """
    Compute variance recursion for GARCH and related models

    Parameters
    ----------
    parameters : 1-d array, float64
        Model parameters
    fresids : 1-d array, float64
        Absolute value of residuals raised to the power in the model.  For
        example, in a standard GARCH model, the power is 2.0.
    sresids : 1-d array, float64
        Variable containing the sign of the residuals (-1.0, 0.0, 1.0)
    sigma2 : 1-d array, float64
        Conditional variances with same shape as resids
    p : int
        Number of symmetric innovations in model
    o : int
        Number of asymmetric innovations in model
    q : int
        Number of lags of the (transformed) variance in the model
    nobs : int
        Length of resids
    backcast : float64
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        transformed variances for each time period
    """

    cdef Py_ssize_t t
    cdef int j, loc

    for t in range(nobs):
        loc = 0
        sigma2[t] = parameters[loc]
        loc += 1
        for j in range(p):
            if (t - 1 - j) < 0:
                sigma2[t] += parameters[loc] * backcast
            else:
                sigma2[t] += parameters[loc] * fresids[t - 1 - j]
            loc += 1
        for j in range(o):
            if (t - 1 - j) < 0:
                sigma2[t] += parameters[loc] * 0.5 * backcast
            else:
                sigma2[t] += parameters[loc] * fresids[t - 1 - j] * (sresids[t-1-j] < 0)
            loc += 1
        for j in range(q):
            if (t - 1 - j) < 0:
                sigma2[t] += parameters[loc] * backcast
            else:
                sigma2[t] += parameters[loc] * sigma2[t - 1 - j]
            loc += 1
        bounds_check(&sigma2[t], &var_bounds[t, 0])

    return np.asarray(sigma2)

def egarch_recursion(double[::1] parameters,
                     double[::1] resids,
                     double[::1] sigma2,
                     int p,
                     int o,
                     int q,
                     int nobs,
                     double backcast,
                     double[:, ::1] var_bounds,
                     double[::1] lnsigma2,
                     double[::1] std_resids,
                     double[::1] abs_std_resids):
    """
    Compute variance recursion for EGARCH models

    Parameters
    ----------
    parameters : 1-d array, float64
        Model parameters
    resids : 1-d array, float64
        Residuals to use in the recursion
    sigma2 : 1-d array, float64
        Conditional variances with same shape as resids
    p : int
        Number of symmetric innovations in model
    o : int
        Number of asymmetric innovations in model
    q : int
        Number of lags of the (transformed) variance in the model
    nobs : int
        Length of resids
    backcast : float64
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        variances for each time period
    lnsigma2 : 1-d array, float64
        Temporary array (overwritten) with same shape as resids
    std_resids : 1-d array, float64
        Temporary array (overwritten) with same shape as resids
    abs_std_resids : 1-d array, float64
        Temporary array (overwritten) with same shape as resids
    """

    cdef double norm_const = 0.79788456080286541  # E[abs(e)], e~N(0,1)
    cdef Py_ssize_t t
    cdef int j, loc

    for t in range(nobs):
        loc = 0
        lnsigma2[t] = parameters[loc]
        loc += 1
        for j in range(p):
            if (t - 1 - j) >= 0:
                lnsigma2[t] += parameters[loc] * (abs_std_resids[t - 1 - j] - norm_const)
            loc += 1
        for j in range(o):
            if (t - 1 - j) >= 0:
                lnsigma2[t] += parameters[loc] * std_resids[t - 1 - j]
            loc += 1
        for j in range(q):
            if (t - 1 - j) < 0:
                lnsigma2[t] += parameters[loc] * backcast
            else:
                lnsigma2[t] += parameters[loc] * lnsigma2[t - 1 - j]
            loc += 1
        if lnsigma2[t] > LNSIGMA_MAX:
            lnsigma2[t] = LNSIGMA_MAX
        sigma2[t] = exp(lnsigma2[t])
        if sigma2[t] < var_bounds[t, 0]:
            sigma2[t] = var_bounds[t, 0]
        elif sigma2[t] > var_bounds[t, 1]:
            sigma2[t] = var_bounds[t, 1] + log(sigma2[t]) - log(var_bounds[t, 1])
        std_resids[t] = resids[t] / sqrt(sigma2[t])
        abs_std_resids[t] = fabs(std_resids[t])

    return np.asarray(sigma2)



def midas_recursion(double[::1] parameters,
                    double[::1] weights,
                    double[::1] resids,
                    double[::1] sigma2,
                    int nobs,
                    double backcast,
                    double[:, ::1] var_bounds):
    """
    Parameters
    ----------
    parameters : 1-d array, float64
        Model parameters
    weights : 1-d array, float64
        Weights for MIDAS recursions
    resids : 1-d array, float64
        Residuals to use in the recursion
    sigma2 : 1-d array, float64
        Conditional variances with same shape as resids
    nobs : int
        Length of resids
    backcast : float64
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        variances for each time period
    """
    cdef Py_ssize_t m, t, i, num_lags
    cdef int j
    cdef double param, omega, alpha, gamma
    cdef double [::1] aw, gw, resids2

    m = weights.shape[0]
    omega = parameters[0]
    alpha = parameters[1]
    gamma = parameters[2]

    aw = np.zeros(m, dtype=np.float64)
    gw = np.zeros(m, dtype=np.float64)
    for i in range(m):
        aw[i] = alpha * weights[i]
        gw[i] = gamma * weights[i]

    resids2 = np.zeros(nobs, dtype=np.float64)

    for t in range(nobs):
        resids2[t] = resids[t] * resids[t]
        sigma2[t] = omega
        for i in range(m):
            if (t - i - 1) >= 0:
                sigma2[t] += (aw[i] + gw[i] * (resids[t - i - 1] < 0)) * resids2[t - i - 1]
            else:
                sigma2[t] +=  (aw[i] + 0.5 * gw[i]) * backcast

        bounds_check(&sigma2[t], &var_bounds[t, 0])

    return np.asarray(sigma2)

cdef double[::1] _figarch_weights(double[::1] parameters,
                                  int p,
                                  int q,
                                  int trunc_lag):

    cdef double phi, d, beta
    cdef double [::1] lam, delta
    cdef Py_ssize_t i
    
    phi = parameters[0] if p else 0.0
    d = parameters[1] if p else parameters[0]
    beta = parameters[p + q] if q else 0.0

    # Recursive weight computation
    lam = np.empty(trunc_lag)
    delta = np.empty(trunc_lag)
    lam[0] = phi - beta + d
    delta[0] = d
    for i in range(1, trunc_lag):
        delta[i] = (i - d) / (i + 1) * delta[i - 1]
        lam[i] = beta * lam[i - 1] + (delta[i] - phi * delta[i - 1])

    return lam


def figarch_weights(double[::1] parameters, int p, int q, int trunc_lag):
    return np.asarray(_figarch_weights(parameters, p, q, trunc_lag))


def figarch_recursion(double[::1] parameters,
                      double[::1] fresids,
                      double[::1] sigma2,
                      int p,
                      int q,
                      int nobs,
                      int trunc_lag,
                      double backcast,
                      double[:, ::1] var_bounds):
    cdef Py_ssize_t t, i
    cdef double bc1, bc2, bc_weight, omega, beta, omega_tilde
    cdef double [::1] lam

    omega = parameters[0]
    beta = parameters[1 + p + q] if q else 0.0
    omega_tilde = omega / (1-beta)
    lam = _figarch_weights(parameters[1:], p, q, trunc_lag)
    for t in range(nobs):
        bc_weight = 0.0
        for i in range(t, trunc_lag):
            bc_weight += lam[i]
        sigma2[t] = omega_tilde + bc_weight * backcast
        for i in range(min(t, trunc_lag)):
            sigma2[t] += lam[i] * fresids[t - i - 1]
        bounds_check(&sigma2[t], &var_bounds[t, 0])

    return np.asarray(sigma2)


def aparch_recursion(double[::1] parameters,
                    double[::1] resids,
                    double[::1] abs_resids,
                    double[::1] sigma2,
                    double[::1] sigma_delta,
                    int p,
                    int o,
                    int q,
                    int nobs,
                    double backcast,
                    double[:, ::1] var_bounds):
    """
    Compute variance recursion for Asymmetric Power ARCH

    Parameters
    ----------
    parameters : ndarray
        Model parameters
    resids : ndarray
        Residuals.
    aresids : ndarray
        Absolute value of residuals.
    sigma2 : ndarray
        Conditional variances with same shape as resids
    sigma_delta : ndarray
        Conditional variance to the power delta with same shape as resids
    p : int
        Number of symmetric innovations in model
    o : int
        Number of asymmetric innovations in model
    q : int
        Number of lags of the (transformed) variance in the model
    nobs : int
        Length of resids
    backcast : float
        Value to use when initializing the recursion
    var_bounds : 2-d array
        nobs by 2-element array of upper and lower bounds for conditional
        transformed variances for each time period
    """
    cdef double delta, shock
    cdef Py_ssize_t t, j

    delta = parameters[1 + p + o + q]
    for t in range(nobs):
        sigma_delta[t] = parameters[0]
        for j in range(p):
            if (t - 1 - j) < 0:
                shock = sqrt(backcast)
            else:
                shock = abs_resids[t - 1 - j]
                if o > j:
                    shock -= parameters[1 + p + j] * resids[t - 1 - j]
            sigma_delta[t] += parameters[1 + j] * (shock ** delta)
        for j in range(q):
            if (t - 1 - j) < 0:
                sigma_delta[t] += parameters[1 + p + o + j] * backcast ** (delta / 2.0)
            else:
                sigma_delta[t] += parameters[1 + p + o + j] * sigma_delta[t - 1 - j]
        sigma2[t] = sigma_delta[t] ** (2.0 / delta)
        bounds_check(&sigma2[t], &var_bounds[t, 0])
        sigma_delta[t] = sigma2[t] ** (delta / 2.0)
    return np.asarray(sigma2)
